from tealish.errors import CompileError
from .tealish_builtins import constants
from .langspec import get_active_langspec


lang_spec = get_active_langspec()

structs = {}


def lookup_op(name):
    if name not in lang_spec.ops:
        raise KeyError(f'Op "{name}" does not exist!')
    return lang_spec.ops[name]


def lookup_constant(name):
    if name not in constants:
        raise KeyError(f'Constant "{name}" does not exist!')
    return constants[name]


def get_field_type(namespace, name):
    if "txn" in namespace:
        return lang_spec.txn_fields[name]
    elif namespace == "global":
        return lang_spec.global_fields[name]


def check_arg_types(name, args):
    op = lookup_op(name)
    arg_types = arg_types = op["arg_types"]
    for i, arg in enumerate(args):
        if arg.type != "any" and arg_types[i] != "any" and arg.type != arg_types[i]:
            raise Exception(
                f"Incorrect type {arg.type} for arg {i} of {name}. Expected {arg_types[i]}"
            )


class BaseNode:
    def process(self):
        pass

    def teal(self):
        return self._teal

    def write_teal(self, writer):
        raise NotImplementedError(self)

    def _tealish(self, formatter=None):
        raise NotImplementedError()

    def tealish(self, formatter=None):
        output = self._tealish(formatter)
        if formatter:
            output = formatter(self, output)
        return output

    def get_scope(self):
        scope = {
            "slots": {},
            "consts": {},
            "blocks": {},
            "functions": {},
        }
        for s in self.get_scopes():
            scope["consts"].update(s["consts"])
            scope["blocks"].update(s["blocks"])
            scope["slots"].update(s["slots"])
            scope["functions"].update(s["functions"])
        return scope

    def get_scopes(self):
        scopes = []
        s = self.get_current_scope()
        while True:
            scopes.append(s)
            if s["parent"]:
                s = s["parent"]
            else:
                break
        return scopes

    def get_const(self, name):
        consts = {}
        for s in self.get_scopes():
            consts.update(s["consts"])
        return consts[name]

    def get_slots(self):
        slots = {}
        for s in self.get_scopes():
            slots.update(s["slots"])
        return slots

    def get_var(self, name):
        slots = self.get_slots()
        if name in slots:
            return slots[name]
        else:
            return (None, None)

    def declare_var(self, name, type):
        slot, _ = self.get_var(name)
        if slot is not None:
            raise Exception(f'Redefinition of variable "{name}"')
        scope = self.get_current_scope()
        if "func__" in scope["name"]:
            # If this var is declared in a function then use the global max slot + 1
            # This is to prevent functions using overlapping slots
            slot = self.compiler.max_slot + 1
        else:
            slot = self.find_slot()
        self.compiler.max_slot = max(self.compiler.max_slot, slot)
        scope["slots"][name] = [slot, type]
        return slot

    def del_var(self, name):
        scope = self.get_current_scope()
        if name in scope["slots"]:
            del scope["slots"][name]

    def find_slot(self):
        scope = self.get_current_scope()
        min, max = scope["slot_range"]
        used_slots = [False] * 255
        slots = self.get_slots()
        for k in slots:
            slot = slots[k][0]
            used_slots[slot] = True
        for i, _ in enumerate(used_slots):
            if not used_slots[i]:
                if i >= min and i <= max:
                    return i
        raise Exception("No available slots!")

    def get_blocks(self):
        blocks = {}
        for s in self.get_scopes():
            blocks.update(s["blocks"])
        return blocks

    def get_block(self, name):
        block = self.get_blocks().get(name)
        # if block:
        #     self.used_blocks.add(block.label)
        return block

    def is_descendant_of(self, node_class):
        return self.find_parent(node_class) is not None

    def find_parent(self, node_class):
        p = self.parent
        while p:
            if isinstance(p, node_class):
                return p
            p = p.parent
        return None

    def has_child_node(self, node_class):
        for node in self.nodes:
            if isinstance(node, node_class) or node.has_child_node(node_class):
                return True
        return False

    def get_current_scope(self):
        return self.parent.get_current_scope()

    def check_arg_types(self, name, args):
        try:
            return check_arg_types(name, args)
        except Exception as e:
            raise CompileError(str(e), node=self)

    def get_field_type(self, namespace, name):
        return get_field_type(namespace, name)

    def lookup_op(self, name):
        return lookup_op(name)

    def lookup_func(self, name):
        scope = self.get_scope()
        if name not in scope["functions"]:
            raise KeyError(f'Func "{name}" not declared in current scope')
        return scope["functions"][name]

    def lookup_var(self, name):
        scope = self.get_scope()
        if name not in scope["slots"]:
            raise KeyError(f'Var "{name}" not declared in current scope')
        return scope["slots"][name]

    def lookup_const(self, name):
        scope = self.get_scope()
        if name not in scope["consts"]:
            raise KeyError(f'Const "{name}" not declared in current scope')
        return scope["consts"][name]

    def lookup_constant(self, name):
        return lookup_constant(name)

    def define_struct(self, struct_name, struct):
        structs[struct_name] = struct

    def get_struct(self, struct_name):
        return structs[struct_name]

    @property
    def line_no(self):
        if hasattr(self, "_line_no"):
            return self._line_no
        if hasattr(self, "parent"):
            return self.parent.line_no



from collections import defaultdict
import json
from operator import index
import re
import sys
from tealish import TealishCompiler

filename = sys.argv[1]
lines = open(filename).readlines()

compiler = TealishCompiler(lines)
compiler.parse()
compiler.process()

new_tealish = compiler.nodes[0].tealish(formatter)


if len(sys.argv) == 2:
    output_filename = filename
else:
    output_filename = sys.argv[2]

if output_filename == "-":
    sys.stdout.write(new_tealish)
else:
    with open(output_filename, "w") as f:
        f.write(new_tealish)



class TestInlineTeal(unittest.TestCase):
    def test_pass_simple(self):
        teal = compile_min(
            [
                "teal:",
                "pushint 1",
                "pushint 2",
                "+",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 1",
                "pushint 2",
                "+",
            ],
        )



import re
import textwrap
from collections import defaultdict
from typing import get_type_hints
from .base import BaseNode
from .errors import CompileError, ParseError
from .tx_expressions import parse_expression

LITERAL_INT = r"[0-9]+"
LITERAL_BYTES = r'"(.+)"'
VARIABLE_NAME = r"[a-z_][a-zA-Z0-9_]*"


class Node(BaseNode):
    pattern = ""
    possible_child_nodes = []

    def __init__(self, line, parent=None, compiler=None) -> None:
        self.parent = parent
        self.current_scope = None
        if parent:
            self.current_scope = parent.current_scope
        self.compiler = compiler
        self.line = line
        self._line_no = compiler.line_no if compiler else None
        # self.child_nodes includes nested nodes (e.g. function body or statements within if...else...end)
        self.child_nodes = []
        # self.nodes includes structural nodes and child_nodes (e.g. function args and body, if conditions and child statements)
        self.nodes = []
        self.properties = {}
        try:
            self.matches = re.match(self.pattern, self.line).groupdict()
        except AttributeError:
            raise ParseError(
                f'Pattern ({self.pattern}) does not match for {self} for line "{self.line}"'
            )
        type_hints = get_type_hints(self.__class__)
        for name, expr_class in type_hints.items():
            if name in self.matches:
                try:
                    if self.matches[name] is not None and hasattr(expr_class, "parse"):
                        value = expr_class.parse(
                            self.matches[name], parent=self, compiler=compiler
                        )
                    else:
                        value = self.matches[name]
                    setattr(self, name, value)
                    if isinstance(value, (Node, Expression, BaseNode)):
                        self.nodes.append(value)
                    self.properties[name] = value
                except Exception as e:
                    raise ParseError(str(e) + f" at line {self.compiler.line_no}")

    def add_child(self, node):
        if not isinstance(node, tuple(self.possible_child_nodes)):
            raise ParseError(
                f"Unexpected child node {node} in {self} at line {self.compiler.line_no}!"
            )
        node.parent = self
        if not node.current_scope:
            node.current_scope = self.current_scope
        self.nodes.append(node)
        self.child_nodes.append(node)

    @classmethod
    def consume(cls, compiler, parent):
        line = compiler.consume_line()
        return cls(line, parent=parent, compiler=compiler)

    def write(self, lines):
        self.compiler.write(lines, self.line_no)

    def get_current_scope(self):
        return self.current_scope

    def new_scope(self, name="", slot_range=None):
        parent_scope = self.parent.get_current_scope() if self.parent else None
        self.current_scope = {
            "parent": parent_scope,
            "slots": {},
            "slot_range": slot_range or [0, 200],
            "aliases": {},
            "consts": {},
            "blocks": {},
            "functions": {},
            "name": (parent_scope["name"] + "__" + name)
            if parent_scope and parent_scope["name"]
            else name,
        }

    def __repr__(self):
        name = self.__class__.__name__
        return name


class Expression(Node):
    @classmethod
    def parse(cls, string, parent, compiler):
        return cls(string)

    @classmethod
    def match(cls, string):
        return re.match(cls.pattern, string) is not None


class Literal(Expression):
    pattern = rf"(?P<value>{LITERAL_BYTES}|{LITERAL_INT})$"

    @classmethod
    def parse(cls, string, parent, compiler):
        for expr in [LiteralInt, LiteralBytes]:
            if expr.match(string):
                return expr.parse(string, parent, compiler)
        raise ParseError(f'Cannot parse "{string}" as Literal')


class LiteralInt(Expression):
    pattern = rf"(?P<value>{LITERAL_INT})$"
    value: int

    def write_teal(self, writer):
        writer.write(self, f"pushint {self.value}")

    def type(self):
        return "int"

    def _tealish(self, formatter=None):
        return f"{self.value}"


class LiteralBytes(Expression):
    pattern = rf"(?P<value>{LITERAL_BYTES})$"
    value: str

    def write_teal(self, writer):
        writer.write(self, f"pushbytes {self.value}")

    def type(self):
        return "bytes"

    def _tealish(self, formatter=None):
        return f"{self.value}"


class Name(Expression):
    pattern = rf"(?P<value>{VARIABLE_NAME})$"
    value: str

    def __init__(self, string) -> None:
        self.slot = None
        self._type = None
        super().__init__(string)

    def _tealish(self, formatter=None):
        return f"{self.value}"

    def type(self):
        return self._type


class GenericExpression(Expression):
    @classmethod
    def parse(cls, string, parent, compiler):
        try:
            node = parse_expression(string)
        except Exception:
            raise ParseError(f'Cannot parse "{string}" as Expression')
        node.parent = parent
        node.compiler = compiler
        return node


class Statement(Node):
    @classmethod
    def consume(cls, compiler, parent):
        line = compiler.peek()
        if line.startswith("block "):
            return Block.consume(compiler, parent)
        elif line.startswith("switch "):
            return Switch.consume(compiler, parent)
        elif line.startswith("func "):
            return Func.consume(compiler, parent)
        elif line.startswith("if "):
            return IfStatement.consume(compiler, parent)
        elif line.startswith("while "):
            return WhileStatement.consume(compiler, parent)
        elif line.startswith("for _"):
            return For_Statement.consume(compiler, parent)
        elif line.startswith("for "):
            return ForStatement.consume(compiler, parent)
        elif line.startswith("teal:"):
            return Teal.consume(compiler, parent)
        elif line.startswith("inner_group:"):
            return InnerGroup.consume(compiler, parent)
        elif line.startswith("inner_txn:"):
            return InnerTxn.consume(compiler, parent)
        elif line.startswith("struct "):
            return Struct.consume(compiler, parent)
        else:
            return LineStatement.consume(compiler, parent)


class Program(Node):
    possible_child_nodes = [Statement]

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.new_scope("")

    def get_current_scope(self):
        return self.current_scope

    @classmethod
    def consume(cls, compiler, parent):
        node = Program("", parent=parent, compiler=compiler)
        expect_struct_definition = True
        while True:
            if compiler.peek() is None:
                break
            n = Statement.consume(compiler, node)
            if not expect_struct_definition and isinstance(n, Struct):
                raise ParseError(
                    f"Unexpected Struct definition at line {n.line_no}. Struct definitions should be at the top of the file and only be preceeded by comments."
                )
            if not isinstance(n, (TealVersion, Blank, Comment, Struct)):
                expect_struct_definition = False
            node.add_child(n)
        return node

    def process(self):
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        for n in self.child_nodes:
            n.write_teal(writer)

    def _tealish(self, formatter=None):
        s = ""
        for n in self.child_nodes:
            s += n.tealish(formatter)
        return s


class InlineStatement(Statement):
    pass


class LineStatement(InlineStatement):
    @classmethod
    def consume(cls, compiler, parent):
        line = compiler.consume_line()
        if line.startswith("#pragma"):
            if compiler.line_no != 1:
                raise ParseError(
                    f'Teal version must be specified in the first line of the program: "{line}" at {compiler.line_no}.'
                )
            return TealVersion(line, parent, compiler=compiler)
        elif line.startswith("#"):
            return Comment(line, parent, compiler=compiler)
        elif line == "":
            return Blank(line, parent, compiler=compiler)
        elif line.startswith("const "):
            return Const(line, parent, compiler=compiler)
        elif line.startswith("int "):
            return IntDeclaration(line, parent, compiler=compiler)
        elif line.startswith("bytes "):
            return BytesDeclaration(line, parent, compiler=compiler)
        elif re.match(r"[A-Z][a-zA-Z_0-9]+ [a-zA-Z_0-9]+ = .*", line):
            return StructDeclaration(line, parent, compiler=compiler)
        elif re.match(r"[a-z][a-zA-Z_0-9]+\.[a-z][a-zA-Z_0-9]* = .*", line):
            return StructAssignment(line, parent, compiler=compiler)
        elif line.startswith("jump "):
            return Jump(line, parent, compiler=compiler)
        elif line.startswith("return"):
            return Return(line, parent, compiler=compiler)
        elif " = " in line:
            return Assignment(line, parent, compiler=compiler)
        elif line.startswith("break"):
            return Break(line, parent, compiler=compiler)
        # Statement functions
        elif line.startswith("exit("):
            return Exit(line, parent, compiler=compiler)
        elif line.startswith("assert("):
            return Assert(line, parent, compiler=compiler)
        elif re.match(r"[a-zA-Z_0-9]+\(.*\)", line):
            return FunctionCallStatement(line, parent, compiler=compiler)
        else:
            raise ParseError(
                f'Unexpected line statement: "{line}" at {compiler.line_no}.'
            )


class TealVersion(LineStatement):
    pattern = r"#pragma version (?P<version>\d+)$"
    version: int

    def write_teal(self, writer):
        writer.write(self, f"#pragma version {self.version}")

    def _tealish(self, formatter=None):
        return f"#pragma version {self.version}\n"


class Comment(LineStatement):
    pattern = r"#(?P<comment>.*)$"
    comment: str

    def write_teal(self, writer):
        writer.write(self, f"//{self.comment}")

    def _tealish(self, formatter=None):
        return f"#{self.comment}\n"


class Blank(LineStatement):
    def write_teal(self, writer):
        writer.write(self, "")

    def _tealish(self, formatter=None):
        return "\n"


class Const(LineStatement):
    pattern = r"const (?P<type>\bint\b|\bbytes\b) (?P<name>[A-Z][a-zA-Z0-9_]*) = (?P<expression>.*)$"
    type: str
    name: str
    expression: Literal

    def process(self):
        scope = self.get_current_scope()
        scope["consts"][self.name] = [self.type, self.expression.value]

    def write_teal(self, writer):
        pass

    def _tealish(self, formatter=None):
        s = f"const {self.type} {self.name}"
        if self.expression:
            s += f" = {self.expression.tealish(formatter)}"
        return s + "\n"


class Jump(LineStatement):
    pattern = r"jump (?P<block_name>.*)$"
    block_name: str

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        b = self.get_block(self.block_name)
        writer.write(self, f"b {b.label}")

    def _tealish(self, formatter=None):
        return f"jump {self.block_name}\n"


class Exit(LineStatement):
    pattern = r"exit\((?P<expression>.*)\)$"
    type: str
    name: str
    expression: GenericExpression

    def process(self):
        self.expression.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, self.expression)
        writer.write(self, "return")

    def _tealish(self, formatter=None):
        return f"exit({self.expression.tealish(formatter)})\n"


class FunctionCallStatement(LineStatement):
    pattern = r"(?P<expression>[a-zA-Z_0-9]+\(.*\))$"
    expression: GenericExpression

    def process(self):
        self.expression.process()
        self.name = self.expression.name

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, self.expression)

    def _tealish(self, formatter=None):
        return f"{self.expression.tealish(formatter)}\n"


class Assert(LineStatement):
    pattern = r'assert\((?P<arg>.*?)(, "(?P<message>.*?)")?\)$'
    arg: GenericExpression
    message: str

    def process(self):
        self.arg.process()
        if self.arg.type not in ("int", "any"):
            raise CompileError(
                f"Incorrect type for assert. Expected int, got {self.arg.type} at line {self.line_no}.",
                node=self,
            )
        if self.message:
            self.compiler.error_messages[self.line_no] = self.message

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, self.arg)
        if self.message:
            writer.write(self, f"assert // {self.message}")
        else:
            writer.write(self, "assert")

    def _tealish(self, formatter=None):
        m = f', "{self.message}"' if self.message else ""
        return f"assert({self.arg.tealish(formatter)}{m})\n"


class BytesDeclaration(LineStatement):
    pattern = r"bytes (?P<name>[a-z][a-zA-Z0-9_]*)( = (?P<expression>.*))?$"
    name: Name
    expression: GenericExpression

    def process(self):
        self.name.slot = self.declare_var(self.name.value, "bytes")
        if self.expression:
            self.expression.process()
            if self.expression.type not in ("bytes", "any"):
                raise CompileError(
                    f"Incorrect type for bytes assignment. Expected bytes, got {self.expression.type}",
                    node=self,
                )

    def write_teal(self, writer):
        writer.write(self, f"// {self.line} [slot {self.name.slot}]")
        if self.expression:
            writer.write(self, self.expression)
            writer.write(self, f"store {self.name.slot} // {self.name.value}")

    def _tealish(self, formatter=None):
        s = f"bytes {self.name.tealish(formatter)}"
        if self.expression:
            s += f" = {self.expression.tealish(formatter)}"
        return s + "\n"


class IntDeclaration(LineStatement):
    pattern = r"int (?P<name>[a-z][a-zA-Z0-9_]*)( = (?P<expression>.*))?$"
    name: Name
    expression: GenericExpression

    def process(self):
        self.name.slot = self.declare_var(self.name.value, "int")
        if self.expression:
            self.expression.process()
            if self.expression.type not in ("int", "any"):
                raise CompileError(
                    f"Incorrect type for int assignment. Expected int, got {self.expression.type}",
                    node=self,
                )

    def write_teal(self, writer):
        writer.write(self, f"// {self.line} [slot {self.name.slot}]")
        if self.expression:
            writer.write(self, self.expression)
            writer.write(self, f"store {self.name.slot} // {self.name.value}")

    def _tealish(self, formatter=None):
        s = f"int {self.name.tealish(formatter)}"
        if self.expression:
            s += f" = {self.expression.tealish(formatter)}"
        return s + "\n"


class Assignment(LineStatement):
    pattern = r"(?P<names>([a-z_][a-zA-Z0-9_]*,?\s*)+) = (?P<expression>.*)$"
    names: str
    expression: GenericExpression

    def process(self):
        self.expression.process()
        t = self.expression.type
        types = t if type(t) == list else [t]
        names = [Name(s.strip()) for s in self.names.split(",")]
        self.names = names
        if len(types) != len(names):
            raise CompileError(
                f"Incorrect number of names ({len(names)}) for values ({len(types)}) in assignment",
                node=self,
            )
        for i, name in enumerate(names):
            if name.value != "_":
                # TODO: we have types for vars now. We should somehow make sure the expression is the correct type
                slot, t = self.get_var(name.value)
                if slot is None:
                    raise CompileError(
                        f'Var "{name.value}" not declared in current scope', node=self
                    )
                if not (types[i] == "any" or types[i] == t):
                    raise CompileError(
                        f"Incorrect type for {t} assignment. Expected {t}, got {types[i]}",
                        node=self,
                    )
                name.slot = slot
                name._type = t

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, self.expression)
        for i, name in enumerate(self.names):
            if name.value == "_":
                writer.write(self, "pop // discarding value for _")
            else:
                writer.write(self, f"store {name.slot} // {name.value}")

    def _tealish(self, formatter=None):
        s = f"{', '.join(n.tealish(formatter) for n in self.names)} = {self.expression.tealish(formatter)}\n"
        return s


class Block(Statement):
    possible_child_nodes = [Statement]
    pattern = r"block (?P<name>[a-zA-Z_0-9]+):$"
    name: str

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        scope = self.get_current_scope()
        scope["blocks"][self.name] = self
        self.label = scope["name"] + ("__" if scope["name"] else "") + self.name
        self.new_scope(self.name)

    @classmethod
    def consume(cls, compiler, parent):
        line = compiler.consume_line()
        block = Block(line, parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            block.add_child(Statement.consume(compiler, block))
        return block

    def process(self):
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        writer.write(self, f"// block {self.name}")
        writer.write(self, f"{self.label}:")
        writer.level += 1
        for n in self.child_nodes:
            n.write_teal(writer)
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = f"block {self.name}:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        output += "end\n"
        return output


class SwitchOption(Node):
    pattern = r"(?P<expression>.*): (?P<block_name>.*)"
    expression: GenericExpression
    block_name: str

    def _tealish(self, formatter=None):
        output = f"{self.expression.tealish(formatter)}: {self.block_name}\n"
        return output


class SwitchElse(Node):
    pattern = r"else: (?P<block_name>.*)"
    block_name: str

    def _tealish(self, formatter=None):
        output = f"else: {self.block_name}\n"
        return output


class Switch(InlineStatement):
    possible_child_nodes = [SwitchOption, SwitchElse]
    pattern = r"switch (?P<expression>.*):$"
    expression: GenericExpression

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.options = []
        self.else_ = None

    def add_option(self, node):
        self.options.append(node)
        self.add_child(node)

    def add_else(self, node):
        self.else_ = node
        self.add_child(node)

    @classmethod
    def consume(cls, compiler, parent):
        switch = Switch(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            if compiler.peek().startswith("else:"):
                switch.add_else(
                    SwitchElse(compiler.consume_line(), switch, compiler=compiler)
                )
            else:
                switch.add_option(
                    SwitchOption(compiler.consume_line(), switch, compiler=compiler)
                )
        return switch

    def process(self):
        self.expression.process()
        for i, node in enumerate(self.options):
            node.expression.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        for i, node in enumerate(self.options):
            writer.write(self, self.expression)
            writer.write(self, node.expression)
            writer.write(self, "==")
            b = self.get_block(node.block_name)
            writer.write(self, f"bnz {b.label}")
        if self.else_:
            b = self.get_block(self.else_.block_name)
            writer.write(self, f"b {b.label} // else")
        else:
            writer.write(self, "err // unexpected value")

    def _tealish(self, formatter=None):
        output = f"switch {self.expression.tealish(formatter)}:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        output += "end\n"
        return output


class TealLine(Node):
    def write_teal(self, writer):
        writer.write(self, f"{self.line}")


class Teal(InlineStatement):
    possible_child_nodes = [TealLine]

    @classmethod
    def consume(cls, compiler, parent):
        node = Teal(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            node.add_child(TealLine.consume(compiler, node))
        return node

    def write_teal(self, writer):
        for n in self.child_nodes:
            n.write_teal(writer)

    def _tealish(self, formatter=None):
        output = "teal:\n"
        for n in self.child_nodes:
            output += indent(n.line) + "\n"
        output += "end\n"
        return output


class InnerTxnFieldSetter(InlineStatement):
    pattern = r"(?P<field_name>.*?)(\[(?P<index>\d\d?)\])?: (?P<expression>.*)"
    field_name: str
    index: int
    expression: GenericExpression

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, self.expression)
        writer.write(self, f"itxn_field {self.field_name}")

    def _tealish(self, formatter=None):
        array_index = f"[{self.index}]" if self.index is not None else ""
        output = f"{self.field_name}{array_index}: {self.expression.tealish(formatter)}"
        return output


class InnerTxn(InlineStatement):
    possible_child_nodes = [InnerTxnFieldSetter]

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.group_index = None
        self.group = None

    @classmethod
    def consume(cls, compiler, parent):
        node = InnerTxn(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            elif compiler.peek().startswith("#"):
                compiler.consume_line()
            else:
                node.add_child(
                    InnerTxnFieldSetter(
                        compiler.consume_line(), node, compiler=compiler
                    )
                )
        group = cls.find_parent(node, InnerGroup)
        if group:
            group.inners.append(node)
            node.group_index = len(group.inners) - 1
            node.group = group
        return node

    def process(self):
        self.array_fields = defaultdict(list)
        for i, node in enumerate(self.child_nodes):
            if node.index is not None:
                index = int(node.index)
                n = len(self.array_fields[node.field_name])
                if n == index:
                    self.array_fields[node.field_name].append(node)
                else:
                    raise ParseError(
                        f"Inccorrect field array index {index} (expected {n}) at line {self.compiler.line_no}!"
                    )
            else:
                node.expression.process()
        for a in self.array_fields.values():
            for node in a:
                node.expression.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        if not self.group:
            writer.write(self, "itxn_begin")
        elif self.group_index > 0:
            writer.write(self, "itxn_next")
        writer.level += 1
        for node in self.child_nodes:
            writer.write(self, node)
        writer.level -= 1
        if not self.group:
            writer.write(self, "itxn_submit")
        writer.write(self, "// end inner_txn")

    def _tealish(self, formatter=None):
        output = "inner_txn:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter)) + "\n"
        output += "end\n"
        return output


class InnerGroup(InlineStatement):
    possible_child_nodes = [Statement]

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.inners = []

    @classmethod
    def consume(cls, compiler, parent):
        node = InnerGroup(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek().startswith("end"):
                compiler.consume_line()
                break
            node.add_child(Statement.consume(compiler, node))
        return node

    def process(self):
        for i, node in enumerate(self.nodes):
            node.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.level += 1
        writer.write(self, "itxn_begin")
        for i, node in enumerate(self.child_nodes):
            writer.write(self, node)
        writer.write(self, "itxn_submit")
        writer.level -= 1
        writer.write(self, "// end inner_group")

    def _tealish(self, formatter=None):
        output = "inner_group:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        output += "end\n"
        return output


class IfThen(Node):
    possible_child_nodes = [InlineStatement]

    @classmethod
    def consume(cls, compiler, parent):
        node = IfThen("", parent, compiler=compiler)
        while True:
            if compiler.peek().startswith(("end", "elif", "else:")):
                break
            node.add_child(InlineStatement.consume(compiler, node))
        return node

    def process(self):
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        writer.write(self, "// then:")
        writer.level += 1
        for n in self.child_nodes:
            n.write_teal(writer)
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = ""
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        return output


class Elif(Node):
    possible_child_nodes = [InlineStatement]
    pattern = r"elif ((?P<modifier>not) )?(?P<condition>.*):"
    condition: GenericExpression
    modifier: str

    @classmethod
    def consume(cls, compiler, parent):
        node = Elif(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek().startswith(("end", "elif", "else:")):
                break
            node.add_child(InlineStatement.consume(compiler, node))
        return node

    def process(self):
        self.condition.process()
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, self.condition)
        if self.modifier == "not":
            writer.write(self, f"bnz {self.next_label}")
        else:
            writer.write(self, f"bz {self.next_label}")
        writer.level += 1
        for n in self.child_nodes:
            n.write_teal(writer)
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = f"elif {'not ' if self.modifier else ''}{self.condition.tealish(formatter)}:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        return output


class Else(Node):
    possible_child_nodes = [InlineStatement]
    pattern = r"else:"

    @classmethod
    def consume(cls, compiler, parent):
        node = Else(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek().startswith(("end")):
                break
            node.add_child(InlineStatement.consume(compiler, node))
        return node

    def process(self):
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.level += 1
        for n in self.child_nodes:
            n.write_teal(writer)
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = "else:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        return output


class IfStatement(InlineStatement):
    possible_child_nodes = [IfThen, Elif, Else]
    pattern = r"if ((?P<modifier>not) )?(?P<condition>.*):$"
    condition: GenericExpression
    modifier: str

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.if_then = None
        self.elifs = []
        self.else_ = None
        self.conditional_index = compiler.conditional_count
        compiler.conditional_count += 1
        self.end_label = f"l{self.conditional_index}_end"

    def add_if_then(self, node):
        node.label = ""
        self.if_then = node
        self.add_child(node)

    def add_elif(self, node):
        i = len(self.elifs)
        node.label = f"l{self.conditional_index}_elif_{i}"
        self.elifs.append(node)
        self.add_child(node)

    def add_else(self, node):
        node.label = f"l{self.conditional_index}_else"
        self.else_ = node
        self.add_child(node)

    @classmethod
    def consume(cls, compiler, parent):
        if_statement = IfStatement(compiler.consume_line(), parent, compiler=compiler)
        if_statement.add_if_then(IfThen.consume(compiler, if_statement))
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            elif compiler.peek().startswith("elif "):
                if_statement.add_elif(Elif.consume(compiler, if_statement))
            elif compiler.peek().startswith("else:"):
                if_statement.add_else(Else.consume(compiler, if_statement))
        return if_statement

    def process(self):
        for i, node in enumerate(self.child_nodes[:-1]):
            node.next_label = self.child_nodes[i + 1].label
        if len(self.child_nodes) > 1:
            self.next_label = self.child_nodes[1].label
        else:
            self.next_label = self.end_label
        self.child_nodes[-1].next_label = self.end_label
        self.condition.process()
        self.if_then.process()
        for i, n in enumerate(self.elifs):
            n.process()
        if self.else_:
            n = self.else_
            n.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.level += 1
        writer.write(self, self.condition)
        if self.modifier == "not":
            writer.write(self, f"bnz {self.next_label}")
        else:
            writer.write(self, f"bz {self.next_label}")
        self.if_then.write_teal(writer)
        if self.elifs or self.else_:
            writer.write(self, f"b {self.end_label}")
        for i, n in enumerate(self.elifs):
            writer.write(self, f"{n.label}:")
            n.write_teal(writer)
            if i != (len(self.elifs) - 1) or self.else_:
                writer.write(self, f"b {self.end_label}")
        if self.else_:
            n = self.else_
            writer.write(self, f"{n.label}:")
            n.write_teal(writer)
        writer.write(self, f"{self.end_label}: // end")
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = f"if {'not ' if self.modifier else ''}{self.condition.tealish(formatter)}:\n"
        for n in self.child_nodes:
            output += n.tealish(formatter)
        output += "end\n"
        return output


class Break(LineStatement):
    pattern = r"break$"

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.parent_loop = self.find_parent(WhileStatement)
        if self.parent_loop is None:
            raise ParseError(
                f'"break" should only be used in a while loop! Line {self.line_no}'
            )

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, f"b {self.parent_loop.end_label}")

    def _tealish(self, formatter=None):
        return "break\n"


class WhileStatement(InlineStatement):
    possible_child_nodes = [InlineStatement]
    pattern = r"while ((?P<modifier>not) )?(?P<condition>.*):$"
    condition: GenericExpression
    modifier: str

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.conditional_index = compiler.conditional_count
        compiler.conditional_count += 1
        self.start_label = f"l{self.conditional_index}_while"
        self.end_label = f"l{self.conditional_index}_end"

    @classmethod
    def consume(cls, compiler, parent):
        node = WhileStatement(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            node.add_child(InlineStatement.consume(compiler, node))
        return node

    def process(self):
        self.condition.process()
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, f"{self.start_label}:")
        writer.level += 1
        writer.write(self, self.condition)
        if self.modifier == "not":
            writer.write(self, f"bnz {self.end_label}")
        else:
            writer.write(self, f"bz {self.end_label}")
        for n in self.child_nodes:
            n.write_teal(writer)
        writer.write(self, f"b {self.start_label}")
        writer.write(self, f"{self.end_label}: // end")
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = f"while {'not ' if self.modifier else ''}{self.condition.tealish(formatter)}:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        output += "end\n"
        return output


class ForStatement(InlineStatement):
    possible_child_nodes = [InlineStatement]
    pattern = r"for (?P<var>[a-z_][a-zA-Z0-9_]*) in (?P<start>[a-zA-Z0-9_]+):(?P<end>[a-zA-Z0-9_]+):$"
    var: str
    start: GenericExpression
    end: GenericExpression

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.conditional_index = compiler.conditional_count
        compiler.conditional_count += 1
        self.start_label = f"l{self.conditional_index}_for"
        self.end_label = f"l{self.conditional_index}_end"

    @classmethod
    def consume(cls, compiler, parent):
        node = ForStatement(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            node.add_child(InlineStatement.consume(compiler, node))
        return node

    def process(self):
        self.var_slot = self.declare_var(self.var, "int")
        for n in self.nodes:
            n.process()
        self.del_var(self.var)

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.level += 1
        writer.write(self, self.start)
        writer.write(self, f"store {self.var_slot} // {self.var}")
        writer.write(self, f"{self.start_label}:")
        writer.write(self, f"load {self.var_slot} // {self.var}")
        writer.write(self, self.end)
        writer.write(self, "==")
        writer.write(self, f"bnz {self.end_label}")
        for n in self.child_nodes:
            n.write_teal(writer)
        writer.write(self, f"load {self.var_slot} // {self.var}")
        writer.write(self, "pushint 1")
        writer.write(self, "+")
        writer.write(self, f"store {self.var_slot} // {self.var}")
        writer.write(self, f"b {self.start_label}")
        writer.write(self, f"{self.end_label}: // end")
        self.del_var(self.var)
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = f"for {self.var} in {self.start.tealish(formatter)}:{self.end.tealish(formatter)}:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        output += "end\n"
        return output


class For_Statement(InlineStatement):
    possible_child_nodes = [InlineStatement]
    pattern = r"for _ in (?P<start>[a-zA-Z0-9_]+):(?P<end>[a-zA-Z0-9_]+):$"
    start: GenericExpression
    end: GenericExpression

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        self.conditional_index = compiler.conditional_count
        compiler.conditional_count += 1
        self.start_label = f"l{self.conditional_index}_for"
        self.end_label = f"l{self.conditional_index}_end"

    @classmethod
    def consume(cls, compiler, parent):
        node = For_Statement(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            node.add_child(InlineStatement.consume(compiler, node))
        return node

    def process(self):
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.level += 1
        writer.write(self, self.start)
        writer.write(self, "dup")
        writer.write(self, f"{self.start_label}:")
        writer.write(self, self.end)
        writer.write(self, "==")
        writer.write(self, f"bnz {self.end_label}")
        for n in self.child_nodes:
            n.write_teal(writer)
        writer.write(self, "pushint 1")
        writer.write(self, "+")
        writer.write(self, "dup")
        writer.write(self, f"b {self.start_label}")
        writer.write(self, "pop")
        writer.write(self, f"{self.end_label}: // end")
        writer.level -= 1

    def _tealish(self, formatter=None):
        output = (
            f"for _ in {self.start.tealish(formatter)}:{self.end.tealish(formatter)}:\n"
        )
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        output += "end\n"
        return output


class ArgsList(Expression):
    arg_pattern = r"(?P<arg_name>[a-z][a-z_0-9]*): (?P<arg_type>int|bytes)"
    pattern = rf"(?P<args>({arg_pattern}(, )?)*)"
    args: str

    def __init__(self, string) -> None:
        super().__init__(string)
        self.args = re.findall(self.arg_pattern, string)

    def _tealish(self, formatter=None):
        output = ", ".join([f"{a}: {t}" for (a, t) in self.args])
        return output


class Func(InlineStatement):
    possible_child_nodes = [InlineStatement]
    pattern = r"func (?P<name>[a-zA-Z_0-9]+)\((?P<args>.*)\)(?P<returns>.*):$"
    name: str
    args: ArgsList
    returns: str

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        scope = self.get_current_scope()
        scope["functions"][self.name] = self
        self.label = scope["name"] + "__func__" + self.name
        self.new_scope("func__" + self.name)
        self.returns = list(filter(None, [s.strip() for s in self.returns.split(",")]))
        self.slots = {}

    @classmethod
    def consume(cls, compiler, parent):
        func = Func(compiler.consume_line(), parent, compiler=compiler)
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            func.add_child(InlineStatement.consume(compiler, func))
        last_node = [n for n in func.nodes if type(n) not in {cls, Comment, Blank}][-1]
        if type(last_node) != Return:
            raise ParseError(
                f"func must end with a return statement at line {compiler.line_no}!"
            )
        return func

    def process(self):
        for (name, type) in self.args.args[::-1]:
            self.slots[name] = self.declare_var(name, type)
        for node in self.nodes:
            node.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        writer.write(self, f"{self.label}:")
        for (name, _) in self.args.args[::-1]:
            slot = self.slots[name]
            writer.write(self, f"store {slot} // {name}")
        for node in self.child_nodes:
            node.write_teal(writer)

    def _tealish(self, formatter=None):
        returns = (" " + (", ".join(self.returns))) if self.returns else ""
        output = f"func {self.name}({self.args.tealish(formatter)}){returns}:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter))
        output += "end\n"
        return output


# class ReturnArgsList(Expression):
#     arg_pattern = r"(?P<arg_name>[a-z][a-z_0-9]*): (?P<arg_type>int|bytes)"
#     pattern = rf"(?P<args>({arg_pattern}(, )?)*)"
#     args: str

#     def __init__(self, string) -> None:
#         super().__init__(string)
#         self.args = re.findall(self.arg_pattern, string)

#     def _tealish(self, formatter=None):
#         output = ", ".join([f"{a}: {t}" for (a, t) in self.args])
#         return output


class Return(LineStatement):
    pattern = r"return ?(?P<args>.*?)?$"
    args: str

    def __init__(self, line, parent=None, compiler=None) -> None:
        super().__init__(line, parent, compiler)
        if not self.is_descendant_of(Func):
            raise ParseError(
                f'"return" should only be used in a function! Line {self.line_no}'
            )
        self.args_expressions = []
        if self.args:
            args = split_return_args(self.args)
            for a in args[::-1]:
                arg = a.strip()
                node = GenericExpression.parse(arg, parent, compiler)
                self.args_expressions.append(node)
        self.nodes = self.args_expressions

    def process(self):
        for n in self.nodes:
            n.process()

    def write_teal(self, writer):
        writer.write(self, f"// {self.line}")
        if self.args:
            for expression in self.args_expressions:
                writer.write(self, expression)
        writer.write(self, "retsub")

    def _tealish(self, formatter=None):
        output = "return"
        if self.args_expressions:
            output += f" {', '.join([e.tealish(formatter) for e in self.args_expressions[::-1]])}"
        return output + "\n"


class StructFieldDefinition(InlineStatement):
    pattern = r"(?P<field_name>[a-z][A-Z-a-z0-9_]*): (?P<data_type>[a-z][A-Z-a-z0-9_]+)(\[(?P<data_length>\d+)\])?"
    field_name: str
    data_type: str
    data_length: int

    def process(self):
        self.size = 8 if self.data_type == "int" else int(self.data_length)

    def write_teal(self, writer):
        pass

    def _tealish(self, formatter=None):
        output = f"{self.field_name}: {self.data_type}"
        return output


class Struct(InlineStatement):
    """
    struct Item:
        asset_id: int
        price: int
        royalty: int
        seller: bytes[32]
        royalty_address: bytes[32]
        round: int
    end
    """

    possible_child_nodes = [StructFieldDefinition]
    pattern = r"struct (?P<name>[A-Z][a-zA-Z_0-9]*):$"
    name: str

    @classmethod
    def consume(cls, compiler, parent):
        node = cls(compiler.consume_line(), parent, compiler=compiler)
        if not isinstance(parent, Program):
            raise ParseError(
                f"Unexpected Struct definition at line {node.line_no}. Struct definitions should be at the top of the file and only be preceeded by comments."
            )
        while True:
            if compiler.peek() == "end":
                compiler.consume_line()
                break
            elif compiler.peek().startswith("#"):
                compiler.consume_line()
            else:
                node.add_child(
                    StructFieldDefinition(
                        compiler.consume_line(), node, compiler=compiler
                    )
                )
        return node

    def process(self):
        for n in self.nodes:
            n.process()
        struct = {
            "fields": {},
            "size": 0,
        }
        offset = 0
        for n in self.child_nodes:
            struct["fields"][n.field_name] = {
                "type": n.data_type,
                "size": n.size,
                "offset": offset,
            }
            offset += n.size
        struct["size"] = offset
        self.define_struct(self.name, struct)

    def write_teal(self, writer):
        pass

    def _tealish(self, formatter=None):
        output = f"struct {self.name}:\n"
        for n in self.child_nodes:
            output += indent(n.tealish(formatter)) + "\n"
        output += "end\n"
        return output


class StructDeclaration(LineStatement):
    pattern = r"(?P<struct_name>[A-Z][a-zA-Z0-9_]*) (?P<name>[a-z][a-zA-Z0-9_]*)( = (?P<expression>.*))?$"
    struct_name: str
    name: Name
    expression: GenericExpression

    def process(self):
        self.name.slot = self.declare_var(self.name.value, ("struct", self.struct_name))
        if self.expression:
            self.expression.process()
            if self.expression.type not in ("bytes", "any"):
                raise CompileError(
                    f"Incorrect type for struct assignment. Expected bytes, got {self.expression.type}",
                    node=self,
                )

    def write_teal(self, writer):
        writer.write(self, f"// {self.line} [slot {self.name.slot}]")
        if self.expression:
            writer.write(self, self.expression)
            writer.write(self, f"store {self.name.slot} // {self.name.value}")

    def _tealish(self, formatter=None):
        s = f"{self.struct_name} {self.name.tealish(formatter)}"
        if self.expression:
            s += f" = {self.expression.tealish(formatter)}"
        return s + "\n"


class StructAssignment(LineStatement):
    pattern = r"(?P<name>[a-z][a-zA-Z0-9_]*).(?P<field_name>[a-z][a-zA-Z0-9_]*)( = (?P<expression>.*))?$"
    name: Name
    field_name: str
    expression: GenericExpression

    def process(self):
        self.name.slot, var_type = self.get_var(self.name.value)
        if type(var_type) != tuple:
            raise CompileError(
                f"{self.name.value} is not a struct reference", node=self
            )
        self.object_type, struct_name = var_type

        struct = self.get_struct(struct_name)
        struct_field = struct["fields"][self.field_name]
        self.offset = struct_field["offset"]
        self.size = struct_field["size"]
        self.data_type = struct_field["type"]
        self.expression.process()
        if self.expression.type not in (self.data_type, "any"):
            raise CompileError(
                f"Incorrect type for struct field assignment. Expected {self.data_type}, got {self.expression.type}",
                node=self,
            )

    def write_teal(self, writer):
        if self.object_type == "struct":
            writer.write(self, f"// {self.line} [slot {self.name.slot}]")
            writer.write(self, f"load {self.name.slot} // {self.name.value}")
            writer.write(self, self.expression)
            if self.data_type == "int":
                writer.write(self, "itob")
            writer.write(
                self, f"replace {self.offset} // {self.name.value}.{self.field_name}"
            )
            writer.write(self, f"store {self.name.slot} // {self.name.value}")

    def _tealish(self, formatter=None):
        s = f"{self.struct_name} {self.name.tealish(formatter)}"
        if self.expression:
            s += f" = {self.expression.tealish(formatter)}"
        return s + "\n"


def split_return_args(s):
    parentheses = 0
    quotes = False
    for i in range(len(s)):
        if s[i] == '"':
            quotes = not quotes
        if not quotes:
            if s[i] == "(":
                parentheses += 1
            if s[i] == ")":
                parentheses -= 1
            if parentheses == 0 and s[i] == ",":
                return [s[:i].strip()] + split_return_args(s[i + 1 :].strip())
    return [s]


def indent(s):
    return textwrap.indent(s, "    ")



from .nodes import Program
from .utils import TealishMap


class TealWriter:
    def __init__(self) -> None:
        self.level = 0
        self.output = []
        self.source_map = {}
        self.current_output_line = 1
        self.current_input_line = 1

    def write(self, parent, node_or_teal):
        parent._teal = []
        if hasattr(node_or_teal, "write_teal"):
            node = node_or_teal
            i = len(self.output)
            node.write_teal(self)
            parent._teal += self.output[i:]
        else:
            teal = node_or_teal
            parent._teal.append(teal)
            prefix = "  " * self.level
            self.output.append(prefix + teal)
            if hasattr(parent, "line_no"):
                self.current_input_line = parent.line_no
            self.source_map[self.current_output_line] = self.current_input_line
            parent.teal_line_no = self.current_output_line
            self.current_output_line += 1


class TealishCompiler:
    def __init__(self, source_lines) -> None:
        self.source_lines = source_lines
        self.output = []
        self.source_map = {}
        self.current_output_line = 1
        self.level = 0
        self.line_no = 0
        self.nodes = []
        self.conditional_count = 0
        self.error_messages = {}
        self.max_slot = 0
        self.writer = TealWriter()
        self.processed = False

    def consume_line(self):
        if self.line_no == len(self.source_lines):
            return
        line = self.source_lines[self.line_no].strip()
        self.line_no += 1
        return line

    def peek(self):
        if self.line_no == len(self.source_lines):
            return
        return self.source_lines[self.line_no].strip()

    def write(self, lines=("",), line_no=0):
        prefix = "  " * self.level
        if type(lines) == str:
            lines = [lines]
        for s in lines:
            self.output.append(prefix + s)
            # print(self.current_output_line, self.output[-1])
            self.source_map[self.current_output_line] = line_no
            self.current_output_line += 1

    def parse(self):
        node = Program.consume(self, None)
        self.nodes.append(node)

    def process(self):
        for node in self.nodes:
            node.process()
        self.processed = True

    def compile(self):
        if not self.nodes:
            self.parse()
        if not self.processed:
            self.process()
        for node in self.nodes:
            node.write_teal(self.writer)
        self.source_map = self.writer.source_map
        self.output = self.writer.output
        return self.writer.output

    def traverse(self, node=None, visitor=None):
        if node is None:
            node = self.nodes[0]
        if visitor:
            visitor(node)
        if getattr(node, "nodes", []):
            for n in node.nodes:
                self.traverse(n, visitor)

    def reformat(self, formatter=None):
        if not self.nodes:
            self.parse()
        if not self.processed:
            self.process()
        return self.nodes[0].tealish(formatter)

    def get_map(self):
        map = TealishMap()
        map.teal_tealish = dict(self.source_map)
        map.errors = dict(self.error_messages)
        return map


def compile_program(source):
    source_lines = source.split("\n")
    compiler = TealishCompiler(source_lines)
    teal = compiler.compile()
    return teal, compiler.get_map()


def compile_lines(source_lines):
    compiler = TealishCompiler(source_lines)
    compiler.parse()
    compiler.compile()
    teal_lines = compiler.output
    return teal_lines


def reformat_program(source):
    source_lines = source.split("\n")
    compiler = TealishCompiler(source_lines)
    output = compiler.reformat()
    output = output.strip() + "\n"
    return output



from .base import BaseNode
from .errors import CompileError


class Integer(BaseNode):
    def __init__(self, value, parent=None) -> None:
        self.value = int(value)
        self.type = "int"
        self.parent = parent

    def write_teal(self, writer):
        writer.write(self, f"pushint {self.value}")

    def _tealish(self, formatter=None):
        return f"pushint {self.value}"


class Bytes(BaseNode):
    def __init__(self, value, parent=None) -> None:
        self.value = value
        self.type = "bytes"
        self.parent = parent

    def write_teal(self, writer):
        writer.write(self, f'pushbytes "{self.value}"')

    def _tealish(self, formatter=None):
        return f'"{self.value}"'


class Variable(BaseNode):
    def __init__(self, name, parent=None) -> None:
        self.name = name
        self.parent = parent

    def process(self):
        self.slot, self.type = self.lookup_var(self.name)
        # is it a struct?
        if type(self.type) == tuple:
            if self.type[0] == "struct":
                self.type = "bytes"

    def write_teal(self, writer):
        writer.write(self, f"load {self.slot} // {self.name}")

    def _tealish(self, formatter=None):
        return f"{self.name}"


class Constant(BaseNode):
    def __init__(self, name, parent=None) -> None:
        self.name = name
        self.type = None
        self.parent = parent

    def process(self):
        type, value = None, None
        try:
            # user defined const
            type, value = self.lookup_const(self.name)
        except KeyError:
            try:
                # builtin TEAL constants
                type, value = self.lookup_constant(self.name)
            except KeyError:
                raise CompileError(
                    f'Constant "{self.name}" not declared in scope', node=self
                )
        if type not in ("int", "bytes"):
            raise CompileError(f"Unexpected const type {type}", node=self)
        self.type = type
        self.value = value

    def write_teal(self, writer):
        if self.type == "int":
            writer.write(self, f"pushint {self.value} // {self.name}")
        elif self.type == "bytes":
            writer.write(self, f"pushbytes {self.value} // {self.name}")

    def _tealish(self, formatter=None):
        return f"{self.name}"


class UnaryOp(BaseNode):
    def __init__(self, op, a, parent=None) -> None:
        self.a = a
        self.op = op
        self.nodes = [a]
        self.parent = parent

    def process(self):
        self.a.process()
        self.check_arg_types(self.op, [self.a])
        op = self.lookup_op(self.op)
        self.type = {"B": "bytes", "U": "int", ".": "any"}[op.get("Returns", "")]

    def write_teal(self, writer):
        writer.write(self, self.a)
        writer.write(self, f"{self.op}")

    def _tealish(self, formatter=None):
        return f"{self.op}{self.a.tealish(formatter)}"


class BinaryOp(BaseNode):
    def __init__(self, a, b, op, parent=None) -> None:
        self.a = a
        self.b = b
        self.op = op
        self.nodes = [a, b]
        self.parent = parent

    def process(self):
        self.a.process()
        self.b.process()
        self.check_arg_types(self.op, [self.a, self.b])
        op = self.lookup_op(self.op)
        self.type = {"B": "bytes", "U": "int", ".": "any"}[op.get("Returns", "")]

    def write_teal(self, writer):
        writer.write(self, self.a)
        writer.write(self, self.b)
        writer.write(self, f"{self.op}")

    def _tealish(self, formatter=None):
        return f"{self.a.tealish(formatter)} {self.op} {self.b.tealish(formatter)}"


class Group(BaseNode):
    def __init__(self, expression, parent=None) -> None:
        self.expression = expression
        self.nodes = [expression]
        self.parent = parent

    def process(self):
        self.expression.process()
        self.type = self.expression.type

    def write_teal(self, writer):
        writer.write(self, self.expression)

    def _tealish(self, formatter=None):
        return f"({self.expression.tealish(formatter)})"


class FunctionCall(BaseNode):
    def __init__(self, name, args, parent=None) -> None:
        self.name = name
        self.args = args
        self.parent = parent
        self.type = None
        self.func_call_type = None
        self.nodes = args
        self.immediate_args = ""

    def process(self):
        func = None
        if self.name in ("error", "push", "pop"):
            return self.process_special_call()
        try:
            func = self.lookup_func(self.name)
        except KeyError:
            pass
        if func:
            return self.process_user_defined_func_call(func)
        try:
            func = self.lookup_op(self.name)
        except KeyError:
            pass
        if func:
            return self.process_op_call(func)
        else:
            raise CompileError(f'Unknown function or opcode "{self.name}"', node=self)

    def process_user_defined_func_call(self, func):
        self.func_call_type = "user_defined"
        self.func = func
        self.type = func.returns[0] if len(func.returns) == 1 else func.returns
        for arg in self.args:
            arg.process()

    def write_teal_user_defined_func_call(self, writer):
        for arg in self.args:
            writer.write(self, arg)
        writer.write(self, f"callsub {self.func.label}")

    def process_op_call(self, op):
        self.func_call_type = "op"
        self.op = op
        immediates = self.args[: (op["Size"] - 1)]
        num_args = len(op.get("Args", ""))

        self.args = self.args[(op["Size"] - 1) :]
        if len(self.args) != num_args:
            raise CompileError(f'Expected {num_args} args for {op["Name"]}!', node=self)
        for i, arg in enumerate(self.args):
            arg.process()
        self.check_arg_types(self.name, self.args)
        for i, x in enumerate(immediates):
            if x.__class__.__name__ == "Constant":
                immediates[i] = x.name
            elif x.__class__.__name__ == "Integer":
                immediates[i] = x.value
        self.immediate_args = " ".join(map(str, immediates))
        returns = [
            {"B": "bytes", "U": "int", ".": "any"}[x] for x in op.get("Returns", "")
        ][::-1]
        self.type = returns[0] if len(returns) == 1 else returns

    def process_special_call(self):
        self.func_call_type = "special"
        self.type = "any"
        for arg in self.args:
            arg.process()

    def write_teal_op_call(self, writer):
        for arg in self.args:
            writer.write(self, arg)
        if self.immediate_args:
            writer.write(self, f"{self.name} {self.immediate_args}")
        else:
            writer.write(self, f"{self.name}")

    def write_teal_special_call(self, writer):
        if self.name == "error":
            writer.write(self, "err")
        elif self.name == "push":
            for arg in self.args:
                writer.write(self, arg)
            writer.write(self, "// push")
        elif self.name == "pop":
            writer.write(self, "// pop")

    def write_teal(self, writer):
        if self.func_call_type == "user_defined":
            self.write_teal_user_defined_func_call(writer)
        elif self.func_call_type == "op":
            self.write_teal_op_call(writer)
        elif self.func_call_type == "special":
            self.write_teal_special_call(writer)

    def _tealish(self, formatter=None):
        args = [a.tealish(formatter) for a in self.args]
        if self.immediate_args:
            args = self.immediate_args.split(", ") + args
        return f"{self.name}({', '.join(args)})"


class TxnField(BaseNode):
    def __init__(self, field, parent=None) -> None:
        self.field = field
        self.type = "any"
        self.parent = parent

    def process(self):
        self.type = self.get_field_type("txn", self.field)

    def write_teal(self, writer):
        writer.write(self, f"txn {self.field}")

    def _tealish(self, formatter=None):
        return f"Txn.{self.field}"


class TxnArrayField(BaseNode):
    def __init__(self, field, arrayIndex, parent=None) -> None:
        self.field = field
        self.arrayIndex = arrayIndex
        self.type = "any"
        self.parent = parent

    def process(self):
        self.type = self.get_field_type("txn", self.field)
        if type(self.arrayIndex) != Integer:
            # index is an expression that needs to be evaluated
            self.arrayIndex.process()

    def write_teal(self, writer):
        if type(self.arrayIndex) != Integer:
            writer.write(self, self.arrayIndex)
            writer.write(self, f"txnas {self.field}")
        else:
            # index is a constant
            writer.write(self, f"txna {self.field} {self.arrayIndex.value}")

    def _tealish(self, formatter=None):
        return f"Txn.{self.field}[{self.arrayIndex.tealish(formatter)}]"


class GroupTxnField(BaseNode):
    def __init__(self, field, index, parent=None) -> None:
        self.field = field
        self.index = index
        self.type = "any"
        self.parent = parent

    def process(self):
        self.type = self.get_field_type("gtxn", self.field)
        if type(self.index) != Integer:
            # index is an expression that needs to be evaluated
            self.index.process()

    def write_teal(self, writer):
        if type(self.index) != Integer:
            # index is an expression that needs to be evaluated
            writer.write(self, self.index)
            writer.write(self, f"gtxns {self.field}")
        else:
            # index is a constant
            assert self.index.value >= 0, "Group index < 0"
            assert self.index.value < 16, "Group index > 16"
            writer.write(self, f"gtxn {self.index.value} {self.field}")

    def _tealish(self, formatter=None):
        return f"Gtxn[{self.index.tealish(formatter)}].{self.field}"


class GroupTxnArrayField(BaseNode):
    def __init__(self, field, index, arrayIndex, parent=None) -> None:
        self.field = field
        self.index = index
        self.arrayIndex = arrayIndex
        self.type = "any"
        self.parent = parent

    def process(self):
        self.type = self.get_field_type("gtxn", self.field)
        if type(self.index) != Integer:
            # index is an expression that needs to be evaluated
            self.index.process()
        if type(self.arrayIndex) != Integer:
            self.arrayIndex.process()

    def write_teal(self, writer):
        if type(self.index) != Integer:
            # index is an expression that needs to be evaluated
            writer.write(self, self.index)
            if type(self.arrayIndex) != Integer:
                # arrayIndex is an expression that needs to be evaluated
                writer.write(self, self.arrayIndex)
                writer.write(self, f"gtxnsas {self.field}")
            else:
                # arrayIndex is a constant
                writer.write(self, f"gtxnsa {self.field} {self.arrayIndex.value}")
        else:
            # index is a constant
            assert self.index.value >= 0 and self.index.value < 16
            if type(self.arrayIndex) != Integer:
                # arrayIndex is an expression that needs to be evaluated
                writer.write(self, self.arrayIndex)
                writer.write(self, f"gtxnas {self.index.value} {self.field}")
            else:
                # arrayIndex is a constant
                writer.write(
                    self,
                    f"gtxna {self.index.value} {self.field} {self.arrayIndex.value}",
                )

    def _tealish(self, formatter=None):
        return f"Gtxn[{self.index.tealish(formatter)}].{self.field}[{self.arrayIndex.tealish(formatter)}]"


class PositiveGroupIndex(BaseNode):
    def __init__(self, index, parent=None) -> None:
        self.index = index
        self.type = "int"
        self.parent = parent

    def write_teal(self, writer):
        writer.write(self, "txn GroupIndex")
        writer.write(self, f"pushint {self.index}")
        writer.write(self, "+")

    def _tealish(self, formatter=None):
        return f"+{self.index}"


class NegativeGroupIndex(BaseNode):
    def __init__(self, index, parent=None) -> None:
        self.index = index
        self.type = "int"
        self.parent = parent

    def write_teal(self, writer):
        writer.write(self, "txn GroupIndex")
        writer.write(self, f"pushint {self.index}")
        writer.write(self, "-")

    def _tealish(self, formatter=None):
        return f"-{self.index}"


class GlobalField(BaseNode):
    def __init__(self, field, parent=None) -> None:
        self.field = field
        self.type = "any"
        self.parent = parent

    def process(self):
        self.type = self.get_field_type("global", self.field)

    def write_teal(self, writer):
        writer.write(self, f"global {self.field}")

    def _tealish(self, formatter=None):
        return f"Global.{self.field}"


class InnerTxnField(BaseNode):
    def __init__(self, field, parent=None) -> None:
        self.field = field
        self.type = "any"
        self.parent = parent

    def process(self):
        self.type = self.get_field_type("txn", self.field)

    def write_teal(self, writer):
        writer.write(self, f"itxn {self.field}")

    def _tealish(self, formatter=None):
        return f"Itxn.{self.field}"


class StructField(BaseNode):
    def __init__(self, name, field, parent=None) -> None:
        self.name = name
        self.field = field
        self.type = "any"
        self.parent = parent

    def process(self):
        self.slot, self.type = self.lookup_var(self.name)
        self.object_type, struct_name = self.type
        struct = self.get_struct(struct_name)
        struct_field = struct["fields"][self.field]
        self.offset = struct_field["offset"]
        self.size = struct_field["size"]
        self.data_type = struct_field["type"]
        self.type = self.data_type

    def write_teal(self, writer):
        if self.object_type == "struct":
            writer.write(self, f"load {self.slot} // {self.name}")
            if self.type == "int":
                writer.write(self, f"pushint {self.offset}")
                writer.write(self, f"extract_uint64 // {self.field}")
            else:
                writer.write(self, f"extract {self.offset} {self.size} // {self.field}")
        else:
            raise Exception()

    def _tealish(self, formatter=None):
        return f"{self.name}.{self.field}"


def class_provider(name):
    classes = {
        "Variable": Variable,
        "Constant": Constant,
        "UnaryOp": UnaryOp,
        "BinaryOp": BinaryOp,
        "Group": Group,
        "Integer": Integer,
        "Bytes": Bytes,
        "FunctionCall": FunctionCall,
        "TxnField": TxnField,
        "TxnArrayField": TxnArrayField,
        "GroupTxnField": GroupTxnField,
        "GroupTxnArrayField": GroupTxnArrayField,
        "PositiveGroupIndex": PositiveGroupIndex,
        "NegativeGroupIndex": NegativeGroupIndex,
        "GlobalField": GlobalField,
        "InnerTxnField": InnerTxnField,
        "StructField": StructField,
    }
    return classes.get(name)



class ParseError(Exception):
    pass


class CompileError(Exception):
    def __init__(self, message: str, node=None) -> None:
        self.node = node
        if node and getattr(node, "line_no", None):
            message += f" at line {node.line_no}"
        super().__init__(message)



constants = {
    "NoOp": ["int", 0],
    "OptIn": ["int", 1],
    "CloseOut": ["int", 2],
    "ClearState": ["int", 3],
    "UpdateApplication": ["int", 4],
    "DeleteApplication": ["int", 5],
    "Pay": ["int", 1],
    "Acfg": ["int", 3],
    "Axfer": ["int", 4],
    "Afrz": ["int", 5],
    "Appl": ["int", 6],
}



import unittest
from unittest import expectedFailure

from tealish import (
    compile_lines,
    TealishCompiler,
    TealWriter,
)
from tealish.errors import (
    CompileError,
    ParseError,
)
from tealish.nodes import Node
from tealish.tx_expressions import parse_expression
from tealish.utils import strip_comments


def compile_min(p):
    teal = compile_lines(p)
    min_teal = strip_comments(teal)
    return min_teal


def compile_expression_min(p, **kwargs):
    teal = compile_expression(p, **kwargs)
    min_teal = strip_comments(teal)
    return min_teal


def compile_expression(expression, scope=None):
    parent = Node("")
    parent.new_scope()
    parent.current_scope.update(scope or {})
    node = parse_expression(expression)
    node.parent = parent
    node.process()
    writer = TealWriter()
    node.write_teal(writer)
    return writer.output


class TestTealVersion(unittest.TestCase):
    def test_one_character(self):
        teal = compile_min(["#pragma version 6"])
        self.assertEqual(teal[0], "#pragma version 6")

    def test_multiple_characters(self):
        teal = compile_min(["#pragma version 12"])
        self.assertEqual(teal[0], "#pragma version 12")

    def test_invalid_version(self):
        with self.assertRaises(ParseError):
            compile_min(["#pragma version a"])


class TestFields(unittest.TestCase):
    def test_txn_array_index_0(self):
        teal = compile_expression_min("Txn.Accounts[0]")
        self.assertEqual(teal, ["txna Accounts 0"])

    def test_txn_array_index_expression(self):
        teal = compile_expression_min("Txn.Accounts[1 + 1]")
        self.assertEqual(teal, ["pushint 1", "pushint 1", "+", "txnas Accounts"])

    def test_group_txn_array_index_0(self):
        teal = compile_expression_min("Gtxn[0].Accounts[0]")
        self.assertEqual(teal, ["gtxna 0 Accounts 0"])

    def test_group_txn_array_index_expression(self):
        teal = compile_expression_min("Gtxn[0].Accounts[1 + 1]")
        self.assertEqual(teal, ["pushint 1", "pushint 1", "+", "gtxnas 0 Accounts"])

    def test_group_txn_index_expression_array_index_expression(self):
        teal = compile_expression_min("Gtxn[1 + 1].Accounts[1 + 1]")
        self.assertEqual(
            teal,
            [
                "pushint 1",
                "pushint 1",
                "+",
                "pushint 1",
                "pushint 1",
                "+",
                "gtxnsas Accounts",
            ],
        )

    def test_group_index_0(self):
        teal = compile_expression_min("Gtxn[0].TypeEnum")
        self.assertEqual(teal, ["gtxn 0 TypeEnum"])

    def test_group_index_1(self):
        teal = compile_expression_min("Gtxn[1].TypeEnum")
        self.assertEqual(teal, ["gtxn 1 TypeEnum"])

    def test_group_index_15(self):
        teal = compile_expression_min("Gtxn[15].TypeEnum")
        self.assertEqual(teal, ["gtxn 15 TypeEnum"])

    def test_group_index_negative(self):
        teal = compile_expression_min("Gtxn[-1].TypeEnum")
        self.assertListEqual(
            teal, ["txn GroupIndex", "pushint 1", "-", "gtxns TypeEnum"]
        )

    def test_group_index_positive(self):
        teal = compile_expression_min("Gtxn[+1].TypeEnum")
        self.assertListEqual(
            teal, ["txn GroupIndex", "pushint 1", "+", "gtxns TypeEnum"]
        )

    def test_group_index_var(self):
        teal = compile_expression_min(
            "Gtxn[index].TypeEnum", scope={"slots": {"index": (0, "int")}}
        )
        self.assertListEqual(teal, ["load 0", "gtxns TypeEnum"])

    def test_group_index_expression(self):
        teal = compile_expression_min("Gtxn[1 + 2].TypeEnum")
        self.assertListEqual(teal, ["pushint 1", "pushint 2", "+", "gtxns TypeEnum"])


class TestIF(unittest.TestCase):
    def test_pass_simple_if(self):
        teal = compile_min(
            [
                "if 1:",
                "exit(0)",
                "end",
            ]
        )
        self.assertListEqual(
            teal, ["pushint 1", "bz l0_end", "pushint 0", "return", "l0_end:"]
        )

    def test_pass_if_not(self):
        teal = compile_min(
            [
                "if not 1:",
                "exit(0)",
                "end",
            ]
        )
        self.assertListEqual(
            teal, ["pushint 1", "bnz l0_end", "pushint 0", "return", "l0_end:"]
        )

    def test_pass_if_else(self):
        teal = compile_min(
            [
                "if 1:",
                "exit(0)",
                "else:",
                "exit(1)",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 1",
                "bz l0_else",
                "pushint 0",
                "return",
                "b l0_end",
                "l0_else:",
                "pushint 1",
                "return",
                "l0_end:",
            ],
        )

    def test_pass_if_elif(self):
        teal = compile_min(
            [
                "if 1:",
                "exit(0)",
                "elif 2:",
                "exit(1)",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 1",
                "bz l0_elif_0",
                "pushint 0",
                "return",
                "b l0_end",
                "l0_elif_0:",
                "pushint 2",
                "bz l0_end",
                "pushint 1",
                "return",
                "l0_end:",
            ],
        )

    def test_pass_if_elif_not(self):
        teal = compile_min(
            [
                "if 1:",
                "exit(0)",
                "elif not 2:",
                "exit(1)",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 1",
                "bz l0_elif_0",
                "pushint 0",
                "return",
                "b l0_end",
                "l0_elif_0:",
                "pushint 2",
                "bnz l0_end",
                "pushint 1",
                "return",
                "l0_end:",
            ],
        )


class TestAssignment(unittest.TestCase):
    def test_assign(self):
        teal = compile_min(["int x = 1"])
        self.assertListEqual(teal, ["pushint 1", "store 0"])

    def test_declare_assign(self):
        teal = compile_min(["int x", "x = 1"])
        self.assertListEqual(teal, ["pushint 1", "store 0"])

    def test_double_assign(self):
        teal = compile_min(
            [
                "int exists",
                "int balance",
                "exists, balance = asset_holding_get(AssetBalance, 0, 123)",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 0",
                "pushint 123",
                "asset_holding_get AssetBalance",
                "store 0",
                "store 1",
            ],
        )

    def test_fail_assign_without_declare(self):
        with self.assertRaises(CompileError) as e:
            _ = compile_min(["x = 1"])
        self.assertEqual(
            e.exception.args[0], 'Var "x" not declared in current scope at line 1'
        )

    def test_fail_invalid(self):
        with self.assertRaises(ParseError):
            compile_min(
                ["int balanceexists, balance = asset_holding_get(AssetBalance, 0, 123)"]
            )


class TestAssert(unittest.TestCase):
    def test_pass_assert_with_message(self):
        teal = compile_min(['assert(1, "Error 1")'])
        self.assertListEqual(teal, ["pushint 1", "assert"])

    def test_pass_assert_with_message_collection(self):
        compiler = TealishCompiler(["assert(0)", 'assert(1, "Error 1")'])
        compiler.compile()
        self.assertDictEqual(compiler.error_messages, {2: "Error 1"})

    def test_pass_simple_assert(self):
        teal = compile_min(["assert(1)"])
        self.assertListEqual(teal, ["pushint 1", "assert"])

    def test_pass_assert_with_group_expression(self):
        compile_lines(["assert(1 && (2 >= 1))"])

    def test_pass_1(self):
        teal = compile_min(["int x = balance(0)"])
        self.assertListEqual(teal, ["pushint 0", "balance", "store 0"])

    def test_fail_wrong_type(self):
        with self.assertRaises(CompileError) as e:
            compile_min(['assert("abc")'])
        self.assertIn(
            "Incorrect type for assert. Expected int, got bytes", str(e.exception)
        )


class TestFunctionReturn(unittest.TestCase):
    def test_pass(self):
        compile_lines(
            [
                "func f():",
                "return",
                "end",
            ]
        )

    def test_fail_no_return(self):
        with self.assertRaises(ParseError) as e:
            compile_lines(
                [
                    "func f():",
                    "assert(1)",
                    "end",
                ]
            )
        self.assertIn("func must end with a return statement", e.exception.args[0])

    @expectedFailure
    def test_fail_wrong_sig_1_return(self):
        with self.assertRaises(CompileError) as e:
            compile_lines(
                [
                    "func f():",
                    "return 1",
                    "end",
                ]
            )
        self.assertIn(
            "Function signature and return statement differ", e.exception.args[0]
        )

    @expectedFailure
    def test_fail_wrong_sig_2_returns(self):
        with self.assertRaises(CompileError) as e:
            compile_lines(
                [
                    "func f() int:",
                    "return 1, 2",
                    "end",
                ]
            )
        self.assertIn(
            "Function signature and return statement differ", e.exception.args[0]
        )

    def test_pass_return_literal(self):
        compile_lines(
            [
                "func f() int:",
                "return 1",
                "end",
            ]
        )

    def test_pass_return_two_literals(self):
        compile_lines(
            [
                "func f() int, int:",
                "return 1, 2",
                "end",
            ]
        )

    def test_pass_return_math_expression(self):
        compile_lines(
            [
                "func f() int:",
                "return 1 + 2",
                "end",
            ]
        )

    def test_pass_return_two_math_expressions(self):
        compile_lines(
            [
                "func f() int, int:",
                "return 1 + 2, 3 + 1",
                "end",
            ]
        )

    def test_pass_return_bytes_with_comma(self):
        teal = compile_min(
            [
                "func f() bytes:",
                'return "1,2,3"',
                "end",
            ]
        )
        self.assertListEqual(teal[1:], ['pushbytes "1,2,3"', "retsub"])

    def test_pass_return_two_func_calls(self):
        teal = compile_min(
            [
                "func f() int, int:",
                "return sqrt(25), exp(5, 2)",
                "end",
            ]
        )
        self.assertListEqual(
            teal[1:], ["pushint 5", "pushint 2", "exp", "pushint 25", "sqrt", "retsub"]
        )


class TestTypeCheck(unittest.TestCase):
    def test_debug(self):
        compile_min(
            [
                "assert(2)",
                "assert(1 + 2)",
            ]
        )

    def test_pass_1(self):
        compile_min(
            [
                "assert(1 + 2)",
            ]
        )

    def test_pass_2(self):
        compile_min(["int a = 3", "assert(btoi(itob(1)) + a)"])

    def test_pass_3(self):
        compile_min(["assert(Txn.Sender == Txn.Receiver)"])

    def test_fail_1(self):
        with self.assertRaises(CompileError) as e:
            compile_min(["bytes x = sqrt(25)"])
        self.assertIn(
            "Incorrect type for bytes assignment. Expected bytes, got int",
            str(e.exception),
        )

    def test_fail_2(self):
        with self.assertRaises(CompileError):
            compile_min(['assert(sqrt("abc"))'])

    def test_fail_3(self):
        with self.assertRaises(CompileError):
            compile_min(["assert(itob(1) + 2)"])

    def test_fail_4(self):
        with self.assertRaises(CompileError) as e:
            compile_min(["int x", "x = itob(2)"])
        self.assertIn(
            "Incorrect type for int assignment. Expected int, got bytes",
            str(e.exception),
        )

    def test_fail_5(self):
        with self.assertRaises(CompileError) as e:
            compile_min(["bytes b", "b = 2"])
        self.assertIn(
            "Incorrect type for bytes assignment. Expected bytes, got int",
            str(e.exception),
        )


class TestInnerGroup(unittest.TestCase):
    def test_pass_simple_inner_group(self):
        teal = compile_min(
            [
                "inner_group:",
                "inner_txn:",
                "TypeEnum: Axfer",
                "Fee: 0",
                "AssetReceiver: Txn.Sender",
                "XferAsset: 10",
                "AssetAmount: 2000000",
                "end",
                "inner_txn:",
                "TypeEnum: Appl",
                "Fee: 2000",
                "ApplicationID: 1",
                'ApplicationArgs[0]: "swap"',
                "ApplicationArgs[1]: 30",
                'ApplicationArgs[2]: "fixed-input"',
                "Accounts[0]: Txn.Accounts[1]",
                "Assets[0]: Txn.Assets[0]",
                "Assets[1]: Txn.Assets[1]",
                "end",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "itxn_begin",
                "pushint 4",
                "itxn_field TypeEnum",
                "pushint 0",
                "itxn_field Fee",
                "txn Sender",
                "itxn_field AssetReceiver",
                "pushint 10",
                "itxn_field XferAsset",
                "pushint 2000000",
                "itxn_field AssetAmount",
                "itxn_next",
                "pushint 6",
                "itxn_field TypeEnum",
                "pushint 2000",
                "itxn_field Fee",
                "pushint 1",
                "itxn_field ApplicationID",
                'pushbytes "swap"',
                "itxn_field ApplicationArgs",
                "pushint 30",
                "itxn_field ApplicationArgs",
                'pushbytes "fixed-input"',
                "itxn_field ApplicationArgs",
                "txna Accounts 1",
                "itxn_field Accounts",
                "txna Assets 0",
                "itxn_field Assets",
                "txna Assets 1",
                "itxn_field Assets",
                "itxn_submit",
            ],
        )

    @expectedFailure
    def test_pass_inner_group_with_if(self):
        # TODO: This currently fails because we don't correctly figure out which is the first txn of the group
        teal = compile_min(
            [
                "int asset_id",
                "inner_group:",
                "if asset_id:",
                "inner_txn:",
                "TypeEnum: Axfer",
                "Fee: 0",
                "AssetReceiver: Txn.Sender",
                "XferAsset: asset_id",
                "AssetAmount: 2000000",
                "end",
                "else:",
                "inner_txn:",
                "TypeEnum: Pay",
                "Fee: 0",
                "Receiver: Txn.Sender",
                "XferAsset: asset_id",
                "Amount: 2000000",
                "end",
                "end",
                "inner_txn:",
                "TypeEnum: Appl",
                "Fee: 2000",
                "ApplicationID: 1",
                'ApplicationArgs[0]: "swap"',
                "ApplicationArgs[1]: 30",
                'ApplicationArgs[2]: "fixed-input"',
                "Accounts[0]: Txn.Accounts[1]",
                "Assets[0]: Txn.Assets[0]",
                "Assets[1]: Txn.Assets[1]",
                "end",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "itxn_begin",
                "load 0",
                "bz l0_else",
                "pushint 4",
                "itxn_field TypeEnum",
                "pushint 0",
                "itxn_field Fee",
                "txn Sender",
                "itxn_field AssetReceiver",
                "load 0",
                "itxn_field XferAsset",
                "pushint 2000000",
                "itxn_field AssetAmount",
                "b l0_end",
                "l0_else:",
                "pushint 1",
                "itxn_field TypeEnum",
                "pushint 0",
                "itxn_field Fee",
                "txn Sender",
                "itxn_field Receiver",
                "load 0",
                "itxn_field XferAsset",
                "pushint 2000000",
                "itxn_field Amount",
                "l0_end:",
                "itxn_next",
                "pushint 6",
                "itxn_field TypeEnum",
                "pushint 2000",
                "itxn_field Fee",
                "pushint 1",
                "itxn_field ApplicationID",
                'pushbytes "swap"',
                "itxn_field ApplicationArgs",
                "pushint 30",
                "itxn_field ApplicationArgs",
                'pushbytes "fixed-input"',
                "itxn_field ApplicationArgs",
                "txna Accounts 1",
                "itxn_field Accounts",
                "txna Assets 0",
                "itxn_field Assets",
                "txna Assets 1",
                "itxn_field Assets",
                "itxn_submit",
            ],
        )

    def test_pass_inner_group_with_statement(self):
        teal = compile_min(
            [
                "int a",
                "int b",
                "inner_group:",
                "int c = a + b",
                "inner_txn:",
                "TypeEnum: Axfer",
                "Fee: 0",
                "AssetReceiver: Txn.Sender",
                "XferAsset: c",
                "AssetAmount: 2000000",
                "end",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "itxn_begin",
                "load 0",
                "load 1",
                "+",
                "store 2",
                "pushint 4",
                "itxn_field TypeEnum",
                "pushint 0",
                "itxn_field Fee",
                "txn Sender",
                "itxn_field AssetReceiver",
                "load 2",
                "itxn_field XferAsset",
                "pushint 2000000",
                "itxn_field AssetAmount",
                "itxn_submit",
            ],
        )


class TestOperators(unittest.TestCase):
    def test_binary(self):
        teal = compile_expression_min("1 || 2")
        self.assertEqual(
            teal,
            [
                "pushint 1",
                "pushint 2",
                "||",
            ],
        )

    def test_unary_literal_int(self):
        teal = compile_expression_min("!1")
        self.assertEqual(
            teal,
            [
                "pushint 1",
                "!",
            ],
        )

    def test_unary_literal_bytes(self):
        teal = compile_expression_min('b~"\x00\x00\x00\x00\x00\x00\x00\x00"')
        self.assertEqual(
            teal,
            [
                'pushbytes "\x00\x00\x00\x00\x00\x00\x00\x00"',
                "b~",
            ],
        )

    def test_unary_variable(self):
        teal = compile_expression_min("!x", scope={"slots": {"x": (0, "int")}})
        self.assertEqual(
            teal,
            [
                "load 0",
                "!",
            ],
        )

    def test_unary_functioncall(self):
        teal = compile_expression_min("!sqrt(25)")
        self.assertEqual(
            teal,
            [
                "pushint 25",
                "sqrt",
                "!",
            ],
        )

    def test_unary_group(self):
        teal = compile_expression_min("!(0 || 1)")
        self.assertEqual(
            teal,
            [
                "pushint 0",
                "pushint 1",
                "||",
                "!",
            ],
        )

    def test_binary_with_unary_b(self):
        teal = compile_expression_min("1 || !1")
        self.assertEqual(
            teal,
            [
                "pushint 1",
                "pushint 1",
                "!",
                "||",
            ],
        )

    def test_binary_with_unary_a(self):
        teal = compile_min(["assert(!1 || 1)"])
        teal = compile_expression_min("!1 || 1")
        self.assertEqual(
            teal,
            [
                "pushint 1",
                "!",
                "pushint 1",
                "||",
            ],
        )


class TestWhile(unittest.TestCase):
    def test_pass_simple(self):
        teal = compile_min(
            [
                "int x = 1",
                "while x < 10:",
                "x = x + 1",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 1",
                "store 0",
                "l0_while:",
                "load 0",
                "pushint 10",
                "<",
                "bz l0_end",
                "load 0",
                "pushint 1",
                "+",
                "store 0",
                "b l0_while",
                "l0_end:",
            ],
        )

    def test_pass_while_break(self):
        teal = compile_min(
            [
                "int x = 1",
                "while 1:",
                "x = x + 1",
                "break",
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 1",
                "store 0",
                "l0_while:",
                "pushint 1",
                "bz l0_end",
                "load 0",
                "pushint 1",
                "+",
                "store 0",
                "b l0_end",
                "b l0_while",
                "l0_end:",
            ],
        )

    def test_fail_break_outside_while(self):
        with self.assertRaises(ParseError) as e:
            compile_min(
                [
                    "int x = 1",
                    "break",
                ]
            )
        self.assertIn('"break" should only be used in a while loop!', str(e.exception))


class TestForLoop(unittest.TestCase):
    def test_pass_implicit(self):
        teal = compile_min(
            [
                "for _ in 0:10:",
                'log("a")',
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 0",
                "dup",
                "l0_for:",
                "pushint 10",
                "==",
                "bnz l0_end",
                'pushbytes "a"',
                "log",
                "pushint 1",
                "+",
                "dup",
                "b l0_for",
                "pop",
                "l0_end:",
            ],
        )

    def test_pass_explicit(self):
        teal = compile_min(
            [
                "for i in 0:10:",
                'log("a")',
                "end",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 0",
                "store 0",
                "l0_for:",
                "load 0",
                "pushint 10",
                "==",
                "bnz l0_end",
                'pushbytes "a"',
                "log",
                "load 0",
                "pushint 1",
                "+",
                "store 0",
                "b l0_for",
                "l0_end:",
            ],
        )


class TestStructs(unittest.TestCase):
    def test_pass_definition(self):
        teal = compile_min(
            [
                "struct Item:",
                "   a: int",
                "   b: int",
                "   c: bytes[10]",
                "end",
            ]
        )
        self.assertListEqual(teal, [])

    def test_fail_definition_lowercase_name(self):
        with self.assertRaises(ParseError):
            compile_min(
                [
                    "struct item:",
                    "   a: int",
                    "   b: int",
                    "   c: bytes[10]",
                    "end",
                ]
            )

    def test_fail_definition_uppercase_field_name(self):
        with self.assertRaises(ParseError):
            compile_min(
                [
                    "struct Item:",
                    "   A: int",
                    "   b: int",
                    "   c: bytes[10]",
                    "end",
                ]
            )

    def test_fail_definition_after_statement(self):
        with self.assertRaises(ParseError):
            compile_min(
                [
                    "int x = 1",
                    "struct Item:",
                    "   a: int",
                    "   b: int",
                    "   c: bytes[10]",
                    "end",
                ]
            )

    def test_fail_definition_inside_statement(self):
        with self.assertRaises(ParseError):
            compile_min(
                [
                    "if 1:",
                    "   struct Item:",
                    "       a: int",
                    "       b: int",
                    "       c: bytes[10]",
                    "   end",
                    "end",
                ]
            )

    def test_pass_declaration(self):
        teal = compile_min(
            [
                "struct Item:",
                "   a: int",
                "   b: int",
                "   c: bytes[10]",
                "end",
                "Item item1 = Txn.ApplicationArgs[0]",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "txna ApplicationArgs 0",
                "store 0",
            ],
        )

    def test_pass_int_field_access(self):
        teal = compile_min(
            [
                "struct Item:",
                "   a: int",
                "   b: int",
                "   c: bytes[10]",
                "end",
                "Item item1 = Txn.ApplicationArgs[0]",
                "assert(item1.a)",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "txna ApplicationArgs 0",
                "store 0",
                "load 0",
                "pushint 0",
                "extract_uint64",
                "assert",
            ],
        )

    def test_pass_byte_field_access(self):
        teal = compile_min(
            [
                "struct Item:",
                "   a: int",
                "   b: int",
                "   c: bytes[10]",
                "end",
                "Item item1 = Txn.ApplicationArgs[0]",
                "log(item1.c)",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "txna ApplicationArgs 0",
                "store 0",
                "load 0",
                "extract 16 10",
                "log",
            ],
        )

    def test_pass_byte_field_assignment(self):
        teal = compile_min(
            [
                "struct Item:",
                "   a: int",
                "   b: int",
                "   c: bytes[10]",
                "end",
                "Item item1 = bzero(28)",
                "item1.c = Txn.ApplicationArgs[0]",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 28",
                "bzero",
                "store 0",
                "load 0",
                "txna ApplicationArgs 0",
                "replace 16",
                "store 0",
            ],
        )

    def test_pass_int_field_assignment(self):
        teal = compile_min(
            [
                "struct Item:",
                "   a: int",
                "   b: int",
                "   c: bytes[10]",
                "end",
                "Item item1 = bzero(28)",
                "item1.a = 1",
            ]
        )
        self.assertListEqual(
            teal,
            [
                "pushint 28",
                "bzero",
                "store 0",
                "load 0",
                "pushint 1",
                "itob",
                "replace 0",
                "store 0",
            ],
        )



from algosdk import source_map


def minify_teal(teal_lines):
    source_map = {}
    n = 1
    output = []
    previous_line_is_label = False
    previous_label = None
    label_replacements = {}
    for i, line in enumerate(teal_lines):
        i = i + 1
        line = line.strip()
        if not (not line or line.startswith("//")):
            if line.split("//")[0].strip().endswith(":"):
                label = line.split("//")[0].strip()[:-1]
                # duplicate labels get compressed into 1
                if previous_line_is_label:
                    label_replacements[label] = previous_label
                    continue
                previous_line_is_label = True
                previous_label = label
            else:
                previous_line_is_label = False
            source_map[n] = i
            n += 1
            output.append(line)
    source_map[n] = len(teal_lines) - 1
    for i, line in enumerate(output):
        for k in label_replacements:
            if k in line:
                output[i] = line.replace(k, label_replacements[k])
    return output, source_map


def strip_comments(teal_lines):
    output = []
    for i, line in enumerate(teal_lines):
        line = line.strip()
        if not (not line or line.startswith("//")):
            line = line.split("//")[0].strip()
            output.append(line)
    return output


class TealishMap:
    def __init__(self, map=None) -> None:
        map = map or {}
        self.pc_teal = {int(k): int(v) for k, v in map.get("pc_teal", {}).items()}
        self.teal_tealish = {
            int(k): int(v) for k, v in map.get("teal_tealish", {}).items()
        }
        self.errors = {int(k): v for k, v in map.get("errors", {}).items()}
        self.tealish_teal = {}
        for teal, tealish in self.teal_tealish.items():
            if tealish not in self.tealish_teal:
                self.tealish_teal[tealish] = []
            self.tealish_teal[tealish].append(teal)

    def get_tealish_line_for_pc(self, pc):
        teal_line = self.get_teal_line_for_pc(pc)
        return self.teal_tealish.get(teal_line, None)

    def get_teal_line_for_pc(self, pc):
        return self.pc_teal.get(pc, None)

    def get_teal_lines_for_tealish(self, tealish_line):
        return self.tealish_teal.get(tealish_line, [])

    def get_tealish_line_for_teal(self, teal_line):
        return self.teal_tealish.get(teal_line, None)

    def get_error_for_pc(self, pc):
        tealish_line = self.get_tealish_line_for_pc(pc)
        return self.errors.get(tealish_line, None)

    def update_from_teal_sourcemap(self, sourcemap):
        if type(sourcemap) == dict:
            sourcemap = source_map.SourceMap(sourcemap)
        self.pc_teal = dict(sourcemap.pc_to_line)

    def as_dict(self):
        return {
            "pc_teal": self.pc_teal,
            "teal_tealish": self.teal_tealish,
            "errors": self.errors,
        }



import json


lang_spec = json.load(open("/Users/fergal/Downloads/langspec.json"))
ops = {op["Name"]: op for op in lang_spec["Ops"]}

abc = "ABCDEFGHIJK"


def type_lookup(a):
    return {
        ".": "Any",
        "B": "bytes",
        "U": "int",
        "": "None",
    }[a]


docs = {}

fields = {}
fields["Global"] = list(
    zip(ops["global"]["ArgEnum"], map(type_lookup, ops["global"]["ArgEnumTypes"]))
)
fields["Txn"] = list(
    zip(ops["txn"]["ArgEnum"], map(type_lookup, ops["txn"]["ArgEnumTypes"]))
)
docs["fields"] = fields

docs["ops"] = {}

for op in ops.values():
    # print(op)
    name = op["Name"]
    immediate_args = op["Size"] - 1
    args = op.get("Args", "")
    arg_list = [f"{abc[i]}: {type_lookup(args[i])}" for i in range(len(args))]
    if "ArgEnum" in op:
        arg_list = ["F: field"] + arg_list
    elif immediate_args:
        arg_list = (["i: int"] * immediate_args) + arg_list
    arg_string = ", ".join(arg_list)
    returns = op.get("Returns", "")[::-1]
    ret = ", ".join([type_lookup(returns[i]) for i in range(len(returns))])
    sig = f"{op['Name']}({arg_string}) -> {ret}"

    docs["ops"][name] = {
        "sig": sig,
        "doc": op["Doc"],
        "doc_extra": op.get("DocExtra"),
        # 'fields': list(zip(op.get('ArgEnum', []), map(type_lookup, op.get('ArgEnumTypes', []))))
    }

json.dump(docs, open("/Users/fergal/Downloads/avmdocs.json", "w"), indent=2)



import sys
from tealish import TealishCompiler

lines = open(sys.argv[1]).readlines()


def f(node):
    # line = node.line
    # scope_path = node.get_current_scope()["name"]
    print(node.__class__.__name__)
    # # func_call = node.expression.expression.node
    # # if func_call.name == 'log':
    # # print(scope_path, line)
    # print(scope_path, line)


callers = {}
funcs = []


def visitor(node, func=None):
    if node.__class__.__name__ == "Func":
        func = node.name
        funcs.append(func)
    if node.__class__.__name__ == "FunctionCall":
        print(node, node.name, func)
        callers.setdefault(node.name, [])
        callers[node.name].append(func)
    else:
        print(node)
    if getattr(node, "nodes", []):
        for n in node.nodes:
            visitor(n, func)


compiler = TealishCompiler(lines)
compiler.parse()
# compiler.traverse(visitor=printer)

indent = 0
root = compiler.nodes[0]


def printer(node, level=0):
    indent = " " * level
    print(f"{indent}{node.__class__.__name__}")
    if getattr(node, "nodes", []):
        print(f"{indent}[")
        for n in node.nodes:
            printer(n, level + 1)
        print(f"{indent}]")


printer(root)

output = compiler.compile()
print("\n".join(output))


# visitor(root)

# edges = []
# for f in funcs:
#     for c in callers.get(f, []):
#         edges.append((c, f))

# for e in edges:
#     print(e)



import json
import pathlib
import click
from tealish import compile_program, reformat_program
from tealish.errors import CompileError, ParseError
from tealish.langspec import (
    fetch_langspec,
    get_active_langspec,
    packaged_lang_spec,
    local_lang_spec,
)
from tealish.build import assemble_with_goal, assemble_with_algod


def _build(path, assembler=None, algod_url=None, quiet=False):
    if path.is_dir():
        paths = path.glob("*.tl")
    else:
        paths = [path]
    for path in paths:
        output_path = pathlib.Path(path).parent / "build"
        output_path.mkdir(exist_ok=True)
        filename = pathlib.Path(path).name
        base_filename = filename.replace(".tl", "")

        # Teal
        teal_filename = output_path / f"{base_filename}.teal"
        if not quiet:
            click.echo(f"Compiling {path} to {teal_filename}")
        teal, tealish_map = _compile_program(open(path).read())
        teal_string = "\n".join(teal + [""])
        with open(teal_filename, "w") as f:
            f.write("\n".join(teal + [""]))

        if assembler:
            tok_filename = output_path / f"{base_filename}.teal.tok"
            if assembler == "goal":
                if not quiet:
                    click.echo(
                        f"Assembling {teal_filename} to {tok_filename} using goal"
                    )
                try:
                    bytecode, sourcemap = assemble_with_goal(teal_string)
                except Exception as e:
                    raise click.ClickException(e)
            elif assembler == "algod":
                if not quiet:
                    click.echo(
                        f"Assembling {teal_filename} to {tok_filename} using algod ({algod_url})"
                    )
                try:
                    bytecode, sourcemap = assemble_with_algod(teal_string, algod_url)
                except Exception as e:
                    raise click.ClickException(e)
            elif assembler == "sandbox":
                raise click.ClickException("Sandbox is not supported yet.")
            else:
                raise Exception()
            with open(tok_filename, "wb") as f:
                f.write(bytecode)
            # Source Map
            tealish_map.update_from_teal_sourcemap(sourcemap)
            map_filename = output_path / f"{base_filename}.map.json"
            if not quiet:
                click.echo(f"Writing source map to {map_filename}")
            with open(map_filename, "w") as f:
                f.write(json.dumps(tealish_map.as_dict()).replace("],", "],\n"))


def _compile_program(source):
    try:
        teal, map = compile_program(source)
    except ParseError as e:
        raise click.ClickException(e)
    except CompileError as e:
        raise click.ClickException(e)
    return teal, map


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("--quiet", "-q", is_flag=True, help="Only print warnings and errors")
@click.pass_context
def cli(ctx, quiet):
    "Tealish Compiler & Tools"
    ctx.ensure_object(dict)
    ctx.obj["quiet"] = quiet


@click.command()
@click.argument("path", type=click.Path(exists=True, path_type=pathlib.Path))
@click.pass_context
def compile(ctx, path):
    """Compile .tl to .teal"""
    _build(path, assembler=None, quiet=ctx.obj["quiet"])


@click.command()
@click.argument("path", type=click.Path(exists=True, path_type=pathlib.Path))
@click.option(
    "--algod",
    "assembler",
    flag_value="algod",
    default=True,
    help="Use algod to compile TEAL [default]",
)
@click.option("--goal", "assembler", flag_value="goal", help="Use goal to compile TEAL")
@click.option(
    "--sandbox", "assembler", flag_value="sandbox", help="Use sandbox to compile TEAL"
)
@click.option(
    "--algod-url",
    type=str,
    default="https://testnet-api.algonode.cloud",
    show_default=True,
    help="Algod URL to use for compiling TEAL",
)
@click.pass_context
def build(ctx, path, assembler, algod_url):
    """Compile .tl to .teal & assemble .teal to .tok (bytecode) & output sourcemap"""
    _build(path, assembler=assembler, algod_url=algod_url, quiet=ctx.obj["quiet"])


@click.command()
@click.argument("tealish_file", type=click.File("r+"))
@click.pass_context
def format(ctx, tealish_file):
    """Rewrite .tl file using standard tealish style"""
    input = tealish_file.read()
    try:
        output = reformat_program(input)
    except ParseError as e:
        raise click.ClickException(e)
    tealish_file.seek(0)
    tealish_file.write(output)
    tealish_file.truncate()


@click.command()
@click.argument("tealish_file", type=click.File("r"))
@click.pass_context
def html(ctx, tealish_file):
    """Output HTML of Tealish & Teal source"""
    raise NotImplementedError()


@click.group()
def langspec():
    """Tools to support new Teal versions by updating the langspec file"""
    pass


@click.command()
@click.pass_context
def langspec_update(ctx):
    """Support new Teal opcodes by updating the langspec.json file from go-algorand master branch"""
    ctx.invoke(langspec_fetch, url_or_branch="master")


@click.command()
@click.argument("url_or_branch", type=str)
@click.pass_context
def langspec_fetch(ctx, url_or_branch):
    """Fetch a specific langspec.json file and use it for the current project. Can be a URL or branch name of go-algorand"""
    new_langspec = fetch_langspec(url_or_branch)
    with open("langspec.json", "w") as f:
        json.dump(new_langspec.as_dict(), f)

    new_ops = new_langspec.new_ops(packaged_lang_spec)
    if new_ops:
        click.echo(f"New ops @ {url_or_branch}:")
    for op in new_ops:
        sig = new_langspec.ops[op]["sig"]
        click.echo(f"{sig}")


@click.command()
@click.argument("url", type=str, default="")
def langspec_diff(url):
    """Show the differences between the current local langspec.json file and the one packaged with this version Tealish"""

    if url:
        local_name = url
        base_langspec = get_active_langspec()
        new_langspec = fetch_langspec(url)
    else:
        local_name = "./langspec.json"
        base_langspec = packaged_lang_spec
        new_langspec = local_lang_spec

    new_ops = new_langspec.new_ops(base_langspec)
    if new_ops:
        click.echo(f"New ops @ {local_name}:")
    for op in new_ops:
        sig = new_langspec.ops[op]["sig"]
        click.echo(f"{sig}")


langspec.add_command(langspec_update, "update")
langspec.add_command(langspec_fetch, "fetch")
langspec.add_command(langspec_diff, "diff")

cli.add_command(compile)
cli.add_command(build)
cli.add_command(format)
cli.add_command(html)
cli.add_command(langspec)



from base64 import b64decode
import json
import subprocess
from algosdk.v2client.algod import AlgodClient
from algosdk.source_map import SourceMap


def assemble_with_goal(teal):
    tmp_out_filename = "/tmp/out.tok"
    try:
        subprocess.check_output(
            ["goal", "clerk", "compile", "-", "--map", "-o", tmp_out_filename],
            input=teal.encode(),
        )
    except FileNotFoundError:
        raise Exception("goal not found in path")
    except subprocess.CalledProcessError as e:
        raise Exception(e.output)
    bytecode = open(tmp_out_filename, "rb").read()
    algod_sourcemap = json.load(open(tmp_out_filename + ".map"))
    return bytecode, SourceMap(algod_sourcemap)


def assemble_with_algod(teal, algod_url):
    token = ""
    if "#" in algod_url:
        algod_url, token = algod_url.split("#")
    algod = AlgodClient(token, algod_url)
    result = algod.compile("\n".join(teal), source_map=True)
    bytecode = b64decode(result["result"])
    algod_sourcemap = result["sourcemap"]
    return bytecode, SourceMap(algod_sourcemap)



import importlib
import os
import requests
import tealish
import json

abc = "ABCDEFGHIJK"
local_lang_spec = None
packaged_lang_spec = None


def type_lookup(a):
    return {
        ".": "any",
        "B": "bytes",
        "U": "int",
        "": "None",
    }[a]


def get_active_langspec():
    return local_lang_spec or packaged_lang_spec


def get_new_local_ops(langspec=None):
    langspec = langspec or local_lang_spec
    if langspec is None:
        return None
    _, new_ops = compare_langspecs(packaged_lang_spec, langspec)
    return new_ops


def compare_langspecs(a, b):
    a_new = []
    b_new = []
    for op in a.ops:
        if op not in b.ops:
            a_new.append(op)
    for op in b.ops:
        if op not in a.ops:
            b_new.append(op)
    return a_new, b_new


def fetch_langspec(url):
    if "http" not in url:
        # assume branch name for go-algorand
        branch = url
        url = f"https://github.com/algorand/go-algorand/raw/{branch}/data/transactions/logic/langspec.json"
    if "github.com" in url and "blob" in url:
        url = url.replace("blob", "raw")
    r = requests.get(url)
    r.raise_for_status()
    langspec_dict = r.json()
    return LangSpec(langspec_dict)


class LangSpec:
    def __init__(self, spec) -> None:
        self.is_packaged = False
        self.spec = spec
        self.fields = {
            "Global": {},
            "Txn": {},
        }
        self.global_fields = self.fields["Global"]
        self.txn_fields = self.fields["Txn"]
        self.ops = {op["Name"]: dict(op) for op in spec["Ops"]}
        for i, field in enumerate(self.ops["global"]["ArgEnum"]):
            self.fields["Global"][field] = type_lookup(
                self.ops["global"]["ArgEnumTypes"][i]
            )
        for i, field in enumerate(self.ops["txn"]["ArgEnum"]):
            self.fields["Txn"][field] = type_lookup(self.ops["txn"]["ArgEnumTypes"][i])

        for op in self.ops.values():
            # print(op)
            name = op["Name"]
            immediate_args = op["Size"] - 1
            args = op.get("Args", "")
            arg_types = [type_lookup(x) for x in args]
            op["arg_types"] = arg_types
            arg_list = [f"{abc[i]}: {arg_types[i]}" for i in range(len(args))]
            if "ArgEnum" in op:
                arg_list = ["F: field"] + arg_list
            elif immediate_args:
                arg_list = (["i: int"] * immediate_args) + arg_list
            arg_string = ", ".join(arg_list)
            returns = op.get("Returns", "")[::-1]
            ret = ", ".join([type_lookup(returns[i]) for i in range(len(returns))])
            sig = f"{name}({arg_string})"
            if ret:
                sig += f"-> {ret}"
            op["sig"] = sig

    def as_dict(self):
        return self.spec

    def new_ops(self, old_spec):
        _, new_ops = compare_langspecs(old_spec, self)
        return new_ops


packaged_lang_spec = LangSpec(
    json.loads(importlib.resources.read_text(package=tealish, resource="langspec.json"))
)
packaged_lang_spec.is_packaged = True
if os.path.exists("langspec.json"):
    local_lang_spec = LangSpec(json.load(open("langspec.json")))



import importlib.resources
import tealish
from textx.metamodel import metamodel_from_file
from .expression_nodes import class_provider

with importlib.resources.path(tealish, "tealish_expressions.tx") as p:
    tealish_mm = metamodel_from_file(
        p,
        use_regexp_group=True,
        skipws=True,
        ws=" \t",
        debug=False,
        classes=class_provider,
    )


def parse_expression(source):
    node = tealish_mm.model_from_str(source)
    return node



import sys
from tealish import TealishCompiler


def html_formatter(node, output):
    if hasattr(node, "html"):
        output = node.html(output)
    return output


def cli():
    filename = sys.argv[1]
    compiler = TealishCompiler(open(filename).readlines())
    compiler.compile()
    output = compiler.reformat(html_formatter)

    output = (
        """
    <html>
    <head>
    <style>
    .TxnField {
        background-color: #EFE;
    }
    .Variable {
        background-color: #FEE;
    }
    body {
        font-family: mono;
    }
    #sidebar {
        width: 400px;
        border: 1px solid #ccc;
        position: fixed;
        top: 10px;
        right: 10px;
        padding: 10px;
        background-color: #eeec;
    }
    </style>
    <script src="../test.trace.js"></script>
    </head>
    <body>
    <pre id="sidebar">
    </pre>
    <pre>
    """
        + output
        + """
    </pre>
    <script>
    var step = 0;
    var currentSlots = {};

    function next(){
        step += 1;
        var s = trace.steps[step];
        var span = document.querySelector(`.line[data-line="${s.line_no + 1}"]`)
        span.style.backgroundColor = "#AFA";
        span.querySelectorAll(".Variable[data-slot]").forEach(function(x){
            console.log(x.dataset.name, x.dataset.slot, tealValue(s.prev_scratch[x.dataset.slot]));
        })
        span.querySelectorAll(".assignment[data-slot]").forEach(function(x){
            console.log(x.dataset.name, x.dataset.slot, tealValue(s.scratch[x.dataset.slot]));
            currentSlots[x.dataset.name] = tealValue(s.scratch[x.dataset.slot]);
            updateSlots(currentSlots);
        })
        span.scrollIntoView({behavior: "smooth", block: "center"});

    }

    function updateSlots(data){
        var s = Object.keys(currentSlots).map(k => `${k}: ${currentSlots[k]}`).join("\\n")
        sidebar.innerHTML = s;
    }

    function tealValue(value){
        if(value){
            return value.tt == 2 ? value.ui || 0: value.tb;
        }
    }

    function lookupTealLine(teal_line){
        var i = trace.history.findIndex(function(t){return trace.pc_map[t.pc] == teal_line});
        var stack = trace.history[i + 1].stack;
        var value = stack[stack.length - 1];
        return tealValue(value);
    }

    function lookupStackValueForElement(e){
        var value = lookupTealLine(parseInt(e.dataset.tealLine) - 1);
        updateSlots(currentSlots)
        sidebar.innerHTML += "\\n\\n" + e.innerText + ": " + value;
    }

    function lookupSlotValueForElement(e){
        var slot = parseInt(e.dataset.slot);
        var line = parseInt(e.parentElement.dataset.line);
        var slots = trace.steps.find(s => s.line_no == (line - 1)).scratch;
        var value = tealValue(slots[slot]);
        updateSlots(currentSlots);
        sidebar.innerHTML += "\\n\\n" + e.innerText + ": " + value;
    }

    document.querySelectorAll("[data-teal-line]").forEach(function(e){
        e.onclick = function(event){
            event.stopPropagation();
            lookupStackValueForElement(event.target);
        }
    })

    document.querySelectorAll(".assignment[data-slot]").forEach(function(e){
        e.onclick = function(event){
            event.stopPropagation();
            lookupSlotValueForElement(event.target);
        }
    })
    document.onkeyup = function(e){if(e.key == "n"){next()}};
    </script>
    </body>
    </html>
    """
    )

    if len(sys.argv) == 2:
        output_filename = filename
    else:
        output_filename = sys.argv[2]

    if output_filename == "-":
        print(output)
    else:
        with open(output_filename, "w") as f:
            f.write(output)


if __name__ == "__main__":
    cli()
